import jittor as jt
from jittor import nn

def _weight_norm(v, g, dim):
    return v * (g / jt.norm(v, 2, dim, keepdim=True))

class WeightNorm(object):
    def __init__(self, name: str, dim: int) -> None:
        if dim is None:
            dim = -1
        self.name = name
        self.dim = dim

    # TODO Make return type more specific
    def compute_weight(self, module: nn.Module):
        g = getattr(module, self.name + '_g')
        v = getattr(module, self.name + '_v')
        return _weight_norm(v, g, self.dim)

    @staticmethod
    def apply(module, name: str, dim: int):
        if hasattr(module, '__fhook2__') and isinstance(module.__fhook2__, WeightNorm):
            raise RuntimeError("Cannot register two weight_norm hooks on "
                                "the same parameter {}".format(name))

        if dim is None:
            dim = -1

        fn = WeightNorm(name, dim)

        weight = getattr(module, name)
        # todo: add check
        # remove w from parameter list
        # del module._parameters[name]
        delattr(module, name)

        # add g and v as new parameters and express w as g/||v|| * v
        module.__setattr__(name + '_g', jt.norm(weight, 2, dim, keepdim=True).detach())
        module.__setattr__(name + '_v', weight.detach())
        setattr(module, name, fn.compute_weight(module))

        # recompute weight before every forward()
        # todo: support multiple hook in a module
        module.register_pre_forward_hook(fn)
        return fn

    def remove(self, module: nn.Module) -> None:
        weight = self.compute_weight(module)
        delattr(module, self.name)
        delattr(module, self.name + '_g')
        delattr(module, self.name + '_v')
        setattr(module, self.name, weight.detach())

    def __call__(self, module: nn.Module, inputs) -> None:
        setattr(module, self.name, self.compute_weight(module))

def weight_norm(module, name, dim):
    ''' Add a module weight normalization.

    :param module: input model.
    :param name: name of the assigned parameter.
    :param dim: which dim to carry out weightnorm.

    Example::

    class jt_module(jt.nn.Module):
        def __init__(self, weight):
            super().__init__()
            self.linear = jt.array(weight)

        def execute(self, x):
            return jt.matmul(self.linear, x)
    
    jm = jt_module(weight)
    weight_norm(jm, 'linear', -1)
    
    '''
    WeightNorm.apply(module, name, dim)
    return module

def remove_weight_norm(module, name: str = 'weight'):
    if hasattr(module, "__fhook2__") and isinstance(module.__fhook2__, WeightNorm):
        delattr(module, "__fhook2__")
        return module
    raise ValueError("weight_norm of '{}' not found in {}"
                     .format(name, module))


# ***************************************************************
# Copyright (c) 2023 Jittor. All Rights Reserved. 
# Maintainers:
#     Guoye Yang <498731903@qq.com>
#     Dun Liang <randonlang@gmail.com>.
#
# 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import jittor as jt
from jittor import init
from jittor import nn
from jittor import Function

class DepthwiseConv(Function):
    def __init__(self, stride=1, padding=0, dilation=1):
        self.stride = stride if isinstance(stride, tuple) else (stride, stride)
        self.padding = padding if isinstance(padding, tuple) else (padding, padding)
        self.dilation = dilation if isinstance(dilation, tuple) else (dilation, dilation)

    def execute(self, x, weight):
        if not jt.flags.use_cuda or not jt.compiler.is_cuda:
            return nn.conv2d(x, weight, None, self.stride, self.padding, self.dilation, x.shape[1])
        self.save_vars = x, weight
        N,C,H,W = x.shape
        o,i,Kh,Kw = weight.shape
        assert(o == C)
        oh = (H+self.padding[0]*2-Kh*self.dilation[0]+self.dilation[0]-1)//self.stride[0]+1
        ow = (W+self.padding[1]*2-Kw*self.dilation[1]+self.dilation[1]-1)//self.stride[1]+1
        filter_height, filter_width = Kh, Kw
        self.Khw = Kh, Kw
        assert oh>0 and ow>0
        output = jt.code(
            [N, C, oh, ow],
            x.dtype,
            [x, weight],
            cuda_header = """
        template <typename T, 
            int filter_height,
            int filter_width, 
            int stride_height,
            int stride_width>
        __global__ void KernelDepthwiseConv(
            const T *const input_data, const T *const filter_data, const int batch_size,
            const int output_channels, const int output_height,
            const int output_width, const int input_channels,  
            const int input_height, const int input_width,     
            const int padding_height, const int padding_width, 
            const int dilate_height, const int dilate_width, T *const output_data) {
            const int kWeghtSize = filter_height * filter_width;
            T r_weight[kWeghtSize];
            const int batch = blockIdx.y;
            const int c_out = blockIdx.x;
            const T* weight = filter_data + c_out * filter_height * filter_width;
            for (int i = 0; i < filter_height * filter_width; i++) r_weight[i] = weight[i];

            for (int w_out = threadIdx.x; w_out < output_width; w_out += blockDim.x) {
                for (int h_out = threadIdx.y; h_out < output_height; h_out += blockDim.y) {
                    const int batch = blockIdx.y;
                    const int c_out = blockIdx.x;

                    const int c_in = c_out;
                    T value = 0;
                    const int h_in_start = -padding_height + h_out * stride_height;
                    const int w_in_start = -padding_width + w_out * stride_width;
                    const int h_in_end = h_in_start + filter_height * dilate_height;
                    const int w_in_end = w_in_start + filter_width * dilate_width;

                    const int in_offset =
                        ((batch * input_channels + c_in) * input_height) * input_width;

                    const int h_end = h_in_end < input_height ? h_in_end : input_height;
                    const int w_end = w_in_end < input_width ? w_in_end : input_width;
                    const int h_start = h_in_start > 0 ? h_in_start : 0;
                    const int w_start = w_in_start > 0 ? w_in_start : 0;

                    for (int h_in = h_in_start, h_f = 0; h_f < filter_height;
                        h_in += dilate_height, h_f++) {
                        for (int w_in = w_in_start, w_f = 0; w_f < filter_width;
                            w_in += dilate_width, w_f++) {
                            if (h_in >= 0 && h_in < input_height && w_in >= 0 &&
                                w_in < input_width) {
                                const int offset = in_offset + h_in * input_width + w_in;
                                value += r_weight[h_f * filter_width + w_f] * input_data[offset];
                            }
                        }
                    }
                    int index =
                        ((batch * gridDim.x + c_out) * output_height + h_out) * output_width +
                        w_out;
                    output_data[index] = value;
                }
            }
        }
        """,
        cuda_src=f"""
            @alias(input, in0)
            @alias(filter, in1)
            @alias(output, out)
            
            const int batch_size = input_shape0;
            const int input_channels = input_shape1;
            const int input_height = input_shape2;
            const int input_width = input_shape3;
            const int output_channels = output_shape1;
            const int output_height = output_shape2;
            const int output_width = output_shape3;
            const int ksize_height = {Kh};
            const int ksize_width = {Kw};
            const int stride_height = {self.stride[0]};
            const int stride_width = {self.stride[1]};
            const int padding_height = {self.padding[0]};
            const int padding_width = {self.padding[1]};
            const int dilate_height = {self.dilation[0]};
            const int dilate_width = {self.dilation[1]};

            int thread = 512;
            if (output_width > 1024 && output_width <= 2048)
                thread = (output_width - 1) / 2 + 1;
            else if (output_width > 512 && output_width <= 1024)
                thread = output_width;
            int blocks = std::min(std::max(thread / output_width, 1), output_height);
            dim3 threads(std::min(output_width, thread), blocks, 1);
            dim3 grid(output_channels, batch_size, 1);
            KernelDepthwiseConv<
                input_type, ksize_height, ksize_width, 
                stride_height, stride_width>
            <<<grid, threads>>>( 
                input_p, filter_p, batch_size, output_channels, output_height,
                output_width, input_channels, input_height, input_width,
                padding_height, padding_width, dilate_height,
                dilate_width, output_p);
        """
        )
        return output

    def grad(self, grad):
        x, weight = self.save_vars
        Kh, Kw = self.Khw
        return jt.code([x.shape, weight.shape], [x.dtype, weight.dtype], [x, weight, grad],
        cuda_header = f"#include <{jt.compile_extern.cub_home}cub/cub.cuh>"+"""
    template <typename T>
    __device__ __inline__ void CudaAtomicAddWithWarp(T* sum, T value) {
    typedef cub::WarpReduce<T> WarpReduce;
    typename WarpReduce::TempStorage temp_storage;
    value = WarpReduce(temp_storage).Sum(value);
    if (cub::LaneId() == 0) 
        atomicAdd(sum, value);
    }
    
    // CUDA kernel to compute the depthwise convolution backprop w.r.t input.
    template <typename T, 
        int filter_height,
        int filter_width, 
        int stride_height,
        int stride_width>
    __global__ void KernelDepthwiseConvInputGradCFilter(
        const T *const input_data, const T *const output_grad_data,
        const T *const filter_data, const int batch_size,   
        const int output_channels, const int output_height, 
        const int output_width, const int input_channels,   
        const int input_height, const int input_width,      
        const int padding_height, const int padding_width,  
        const int dilate_height, const int dilate_width,    
        T *const input_grad_data) {
        const int kWeghtSize = filter_height * filter_width + 1;
        T r_weight[kWeghtSize];
        const int batch = blockIdx.y;
        const int c_in = blockIdx.x;

        const T* weight = filter_data + c_in * filter_height * filter_width;
        for (int i = 0; i < filter_height * filter_width; i++)
            r_weight[i] =
                weight[filter_height * filter_width - i - 1];

        for (int w_in = threadIdx.x; w_in < input_width; w_in += blockDim.x) {
            for (int h_in = threadIdx.y; h_in < input_height; h_in += blockDim.y) {
                const int batch = blockIdx.y;
                const int c_in = blockIdx.x;

                int h_out_start = h_in - (filter_height - 1) * dilate_height + padding_height;

                int w_out_start = w_in - (filter_width - 1) * dilate_width + padding_width;

                T value = 0;
                int index =
                    ((batch * gridDim.x + c_in) * input_height + h_in) * input_width +
                    w_in;

                for (int h_out = h_out_start, h_f = 0; h_f < filter_height;
                    h_out += dilate_height, h_f++) {
                    for (int w_out = w_out_start, w_f = 0; w_f < filter_width;
                        w_out += dilate_width, w_f++) {
                        int s_h_out = h_out / stride_height;
                        int s_w_out = w_out / stride_width;
                        if (h_out % stride_height == 0 && w_out % stride_width == 0 &&
                            s_h_out >= 0 && s_h_out < output_height && s_w_out >= 0 &&
                            s_w_out < output_width) {
                        const int output_grad_offset =
                            ((batch * output_channels + c_in) * output_height +
                            s_h_out) *
                                output_width +
                            s_w_out;
                        value +=
                            output_grad_data[output_grad_offset] *
                            r_weight[h_f * filter_width + w_f];
                        }
                    }
                }
                input_grad_data[index] = value;
            }
        }
    }

    // Cuda kernel to compute the depthwise convolution backprop w.r.t. filter.
    template <typename T>
    __global__ void KernelDepthwiseConvFilterGrad(
        const T* output_grad_data, const T* input_data, const int num,
        const int output_channels, const int output_height, const int output_width,
        const int input_channels, const int input_height, const int input_width,
        const int filter_height,
        const int filter_width, const int stride_height, const int stride_width,
        const int padding_height, const int padding_width, const int dilate_height,
        const int dilate_width, T* filter_grad_data) {
        T s = 0;

        int gbid = (((blockIdx.z * blockDim.z + threadIdx.z) * gridDim.y) + blockIdx.y) * gridDim.x + blockIdx.x;

        for (int image_w = threadIdx.x; image_w < output_width;
            image_w += blockDim.x) {
            for (int bid = 0; bid < num; bid++) {
            //for (int bid = threadIdx.z; bid < num; bid+=blockDim.z) {
                for (int image_h = threadIdx.y; image_h < output_height;
                    image_h += blockDim.y) {
                    int kernel_id = blockIdx.z;
                    int kernel_h = blockIdx.y * dilate_height - padding_height;
                    int kernel_w = blockIdx.x * dilate_width - padding_width;

                    int image_hk = image_h * stride_height + kernel_h;
                    int image_wk = image_w * stride_width + kernel_w;
                    if (image_hk < 0 || image_hk >= input_height) continue;
                    if (image_wk < 0 || image_wk >= input_width) continue;
                    #define gaid(N, C, H, W) \
                    ((((N)*gridDim.z + (C)) * output_height + (H)) * output_width + (W))
                            int input_id = ((bid * gridDim.z +
                                            kernel_id) *
                                                input_height +
                                            image_hk) *
                                            input_width +
                                        image_wk;
                            s += output_grad_data[gaid(bid, kernel_id, image_h, image_w)] *
                                input_data[input_id];

                    #undef gaid
                }
            }
        }
        CudaAtomicAddWithWarp(&filter_grad_data[gbid], s);
    }
        """,
    cuda_src=f"""
    // source for backward to data
        @alias(input, in0)
        @alias(filter, in1)
        @alias(output_grad, in2)
        @alias(input_grad, out0)
        @alias(filter_grad, out1)

        const int batch_size = input_shape0;
        const int input_channels = input_shape1;
        const int input_height = input_shape2;
        const int input_width = input_shape3;
        const int output_channels = output_grad_shape1;
        const int output_height = output_grad_shape2;
        const int output_width = output_grad_shape3;
        const int ksize_height = {Kh};
        const int ksize_width = {Kw};
        const int stride_height = {self.stride[0]};
        const int stride_width = {self.stride[1]};
        const int padding_height = {self.padding[0]};
        const int padding_width = {self.padding[1]};
        const int dilate_height = {self.dilation[0]};
        const int dilate_width = {self.dilation[1]};

        int thread = 512;
        if (input_width > 1024 && input_width <= 2048)
        thread = (input_width - 1) / 2 + 1;
        else if (input_width > 512 && input_width <= 1024)
        thread = input_width;
        int blocks = std::min(std::max(thread / input_width, 1), input_height);
        dim3 threads(std::min(input_width, thread), blocks, 1);
        dim3 grid(input_channels, batch_size, 1);
        KernelDepthwiseConvInputGradCFilter<
            input_type, ksize_height, ksize_width
            , stride_height, stride_width>
            <<<grid, threads, 0>>>( 
            input_p, output_grad_p, filter_p, batch_size,          
            output_channels, output_height, output_width, input_channels,   
            input_height, input_width, padding_height,       
            padding_width, dilate_height, dilate_width, input_grad_p);   

    // source for backward to filter
    
        int block_size = 512;
        if (output_width > 1024 && output_width <= 2048)
        block_size = (output_width - 1) / 2 + 1;
        else if (output_width > 512 && output_width <= 1024)
        block_size = output_width;
        int crop_output_height =
            std::min(std::max(block_size / output_width, 1), output_height);

        grid = dim3(ksize_width, ksize_height, output_channels);
        threads = dim3(std::min(output_width, block_size), crop_output_height, 1);
        cudaMemsetAsync(filter_grad_p, 0, filter_grad->size);

        KernelDepthwiseConvFilterGrad<                                         
            input_type><<<grid, threads, 0>>>(      
            output_grad_p, input_p, batch_size, output_channels,           
            output_height, output_width, input_channels, input_height,           
            input_width, ksize_height, ksize_width,           
            stride_height, stride_width, padding_height, padding_width,          
            dilate_height, dilate_width, filter_grad_p);                      
    """
    )


# ***************************************************************
# Copyright (c) 2023 Jittor. All Rights Reserved. 
# Maintainers: Dun Liang <randonlang@gmail.com>. 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************

if __name__ == "__main__":
    import unittest, os
    unittest.TestLoader.sortTestMethodsUsing = None

    suffix = "__main__.py"
    assert __file__.endswith(suffix)
    test_dir = __file__[:-len(suffix)]

    skip_l = int(os.environ.get("test_skip_l", "0"))
    skip_r = int(os.environ.get("test_skip_r", "1000000"))
    skip = os.environ.get("test_skip", "").split(",")
    test_only = None
    if "test_only" in os.environ:
        test_only = set(os.environ.get("test_only").split(","))

    test_files = os.listdir(test_dir)
    test_files = sorted(test_files)
    suite = unittest.TestSuite()
    test_names = []
    seperate_test = os.environ.get("seperate_test", "1") == "1"
    
    for _, test_file in enumerate(test_files):
        test_name = test_file.split(".")[0]
        tests = unittest.defaultTestLoader.loadTestsFromName(
            "jittor.test."+test_name)
            
        if not test_file.startswith("test_"):
            continue
        if _ < skip_l or _ > skip_r:
            continue
        if test_only and test_name not in test_only:
            continue
        for s in skip:
            if s in test_name:
                continue

        print("Add Test", _, test_name)
        if seperate_test:
            test_names.append("jittor.test."+test_name)
        else:
            suite.addTest(tests)

    if seperate_test:
        import subprocess as sp
        import sys
        import time
        import jittor_utils
        start = time.time()
        errors = ""
        f = open(jittor_utils.home()+"/.cache/jittor/test.log", "w")
        for i,test_name in enumerate(test_names):
            progress = f"{i}/{len(test_names)}"
            print(f"[RUN TEST {progress}]", test_name)
            r = sp.run(" ".join([sys.executable, '-m', test_name, '-v']), stdout=sp.PIPE, stderr=sp.STDOUT, timeout=60*10, shell=True)
            out = r.stdout.decode('utf8')
            sys.stdout.write(out)
            f.write(out)
            msg = f"[RUN TEST {progress} OK]"
            if r.returncode:
                msg = f"[RUN TEST {progress} FAILED]"
            msg = msg + f" {test_name} {time.time()-start:.1f}\n"
            if r.returncode:
                errors += msg
            sys.stdout.write(msg)
            f.write(msg)
        sys.stdout.write(errors)
        f.write(errors)
        f.close()

    result = unittest.TextTestRunner(verbosity=3).run(suite)
    if len(result.errors) or len(result.failures):
        exit(1)



#!/usr/bin/python3
# ***************************************************************
# Copyright (c) 2023 Jittor. All Rights Reserved. 
# Maintainers: Dun Liang <randonlang@gmail.com>. 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
from ctypes import cdll
import sys

lib_path = sys.argv[1]
cmd = sys.argv[2]
if not lib_path.endswith(".so"):
    i = -1
    while lib_path[i] != '.':
        i -= 1
    if i > -10: lib_path = lib_path[:i]
    lib_path += ".so"

if cmd == "run_so":
    lib = cdll.LoadLibrary(lib_path)
    lib.fake_main()
    exit(0)

with open(lib_path+".key") as f:
    cpcmd = f.read().splitlines()[0]

def run_cmd(cmd):
    print("Run cmd:", cmd)
    assert os.system(cmd) == 0, "Run cmd failed: "+cmd

import os
if cmd == "cc_to_so":
    run_cmd(cpcmd)
    # remove hash info, force re-compile
    with open(lib_path+'.key', 'w') as f:
        f.write(cpcmd)
elif cmd == "cc_to_s":
    asm_cmd = cpcmd.replace("_op.so", "_op.s") \
        .replace("-g", "") \
        .replace("-lstdc++", "") \
        .replace("-ldl", "") \
        .replace("-shared", "-S")
    run_cmd(asm_cmd)
elif cmd == "s_to_so":
    asm_cmd = cpcmd.replace("_op.cc", "_op.s") \
        .replace(" -g", "")
    run_cmd(asm_cmd)
    # remove hash info, force re-compile
    with open(lib_path+'.key', 'w') as f:
        f.write(cpcmd)
elif cmd == "perf_so":
    perf_cmd = "perf record "+__file__+" "+lib_path+" run_so && perf annotate"
    run_cmd(perf_cmd)
elif cmd == "vtune_so":
    if os.path.isdir("./__res"):
        run_cmd("rm -r ./__res")
    vtune_cmd = "amplxe-cl -collect uarch-exploration -r ./__res "+__file__+" "+lib_path+" run_so"
    run_cmd(vtune_cmd)
else:
    assert 0, "unknown cmd: {cmd}".format(cmd)



# ***************************************************************
# Copyright (c) 2023 Jittor. All Rights Reserved.
# Maintainers:
#     Haoyang Peng <2247838039@qq.com>
#     Guowei Yang <471184555@qq.com>
#     Dun Liang <randonlang@gmail.com>.
#
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import jittor as jt
from jittor import nn
import numpy as np
import unittest
from jittor.weightnorm import weight_norm

try:
    import torch
    from torch.autograd import Variable
    import autograd.numpy as anp
    from autograd import jacobian

    has_autograd = True
except:
    has_autograd = False

class jt_module(jt.nn.Module):
    def __init__(self, weight):
        super().__init__()
        self.linear = jt.array(weight)

    def execute(self, x):
        return jt.matmul(self.linear, x)

class torch_module(torch.nn.Module):
    def __init__(self, weight):
        super().__init__()
        self.linear = torch.nn.Parameter(torch.from_numpy(weight))
    
    def forward(self, x):
        return torch.matmul(self.linear, x)

@unittest.skipIf(not has_autograd, "No autograd found.")
class TestWeightNorm(unittest.TestCase):
    def test_weightnorm(self):
        for i in range(30):
            weight = np.random.uniform(0,1,(i+10,40))
            jm = jt_module(weight)
            tm = torch_module(weight)
            inp = np.random.uniform(0,1,(40,i+30))
            torch.nn.utils.weight_norm(tm, 'linear', -1)
            weight_norm(jm, 'linear', -1)
            jinp = jt.array(inp)
            tinp = Variable(torch.from_numpy(inp), requires_grad=True)
            joup = jm(jinp)
            toup = tm(tinp)
            np.testing.assert_allclose(joup.data, toup.detach().numpy(), rtol=1e-4, atol=1e-6)
            gq = jt.grad(joup, jinp).data
            tgq = torch.autograd.grad(toup, tinp, torch.ones_like(toup), retain_graph=True)
            np.testing.assert_allclose(gq, tgq[0].numpy(), rtol=1e-4, atol=1e-6)

if __name__ == "__main__":
    unittest.main()



# ***************************************************************
# Copyright (c) 2023 Jittor. All Rights Reserved. 
# Maintainers: 
#     Guowei Yang <471184555@qq.com>
#     Dun Liang <randonlang@gmail.com>. 
# 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import os, sys
import jittor as jt
import numpy as np
from jittor.test.test_mpi import run_mpi_test

mpi = jt.compile_extern.mpi
if mpi:
    n = mpi.world_size()

@unittest.skipIf(not jt.in_mpi, "no inside mpirun")
class TestMpiOps(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        np.random.seed(0)
        jt.seed(3)

    def test_all_reduce(self):
        x = jt.random([5, 5])
        y = x.mpi_all_reduce()
        np.testing.assert_allclose(y.data, (x*n).data)
        g = jt.grad(y,x)
        np.testing.assert_allclose(g.data, np.ones([5,5])*n)

    def test_all_reduce_mean(self):
        x = jt.random([5, 5])
        y = x.mpi_all_reduce("mean")
        np.testing.assert_allclose(y.data, x.data)
        g = jt.grad(y,x)
        np.testing.assert_allclose(g.data, np.ones([5,5]))

    def test_broadcast(self):
        data = jt.random([5, 5])
        if mpi.world_rank() == 0:
            x = data
        else:
            x = jt.zeros([5, 5])
        y = x.mpi_broadcast(0)
        np.testing.assert_allclose(y.data, data.data)
        g = jt.grad(y,x)
        if mpi.world_rank() == 0:
            np.testing.assert_allclose(g.data, np.ones([5,5])*n)
        else:
            np.testing.assert_allclose(g.data, np.zeros([5,5]))

    def test_reduce(self):
        x = jt.random([5, 5])
        y = x.mpi_reduce(root=0)
        y.sync()
        if mpi.world_rank() == 0:
            np.testing.assert_allclose(y.data, (x*n).data)
        else:
            np.testing.assert_allclose(y.data, np.zeros([5,5]))
        g = jt.grad(y,x)
        print(mpi.world_rank(), g)
        np.testing.assert_allclose(g.data, np.ones([5,5]))


@unittest.skipIf(not jt.compile_extern.has_mpi, "no mpi found")
class TestMpiOpsEntry(unittest.TestCase):
    def test(self):
        run_mpi_test(2, "test_mpi_op")

if __name__ == "__main__":
    unittest.main()



# ***************************************************************
# Copyright (c) 2023 Jittor. All Rights Reserved. 
# Maintainers: 
#     Wenyang Zhou <576825820@qq.com>
#     Dun Liang <randonlang@gmail.com>. 
# 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import jittor as jt
import numpy as np
import jittor.nn as jnn

skip_this_test = False

try:
    jt.dirty_fix_pytorch_runtime_error()
    import torch
    import torch.nn as tnn
except:
    torch = None
    tnn = None
    skip_this_test = True

def check_equal(arr, j_layer, p_layer):
    jittor_arr = jt.array(arr)
    pytorch_arr = torch.Tensor(arr)
    jittor_result = j_layer(jittor_arr)
    pytorch_result = p_layer(pytorch_arr)
    assert np.allclose(pytorch_result.detach().numpy(), jittor_result.numpy())

@unittest.skipIf(skip_this_test, "No Torch found")
class TestPad(unittest.TestCase):
    def test_pad(self):
        # ***************************************************************
        # Test ReplicationPad2d Layer
        # ***************************************************************
        arr = np.random.randn(16,3,224,224)
        check_equal(arr, jnn.ReplicationPad2d(10), tnn.ReplicationPad2d(10))
        check_equal(arr, jnn.ReplicationPad2d((1,23,4,5)), tnn.ReplicationPad2d((1,23,4,5)))
        check_equal(arr, jnn.ReplicationPad2d((1,0,1,5)), tnn.ReplicationPad2d((1,0,1,5)))
        check_equal(arr, jnn.ReplicationPad2d((100)), tnn.ReplicationPad2d((100)))

        # ***************************************************************
        # Test ConstantPad2d Layer
        # ***************************************************************
        arr = np.random.randn(16,3,224,224)
        check_equal(arr, jnn.ConstantPad2d(10,-2), tnn.ConstantPad2d(10,-2))
        check_equal(arr, jnn.ConstantPad2d((2,3,34,1),10.2), tnn.ConstantPad2d((2,3,34,1),10.2))

        arr = np.random.randn(16,3,224,10,10)
        check_equal(arr, jnn.ConstantPad2d(10,-2), tnn.ConstantPad2d(10,-2))
        check_equal(arr, jnn.ConstantPad2d((2,3,34,1),10.2), tnn.ConstantPad2d((2,3,34,1),10.2))

        # ***************************************************************
        # Test ZeroPad2d Layer
        # ***************************************************************
        arr = np.random.randn(16,3,224,224)
        check_equal(arr, jnn.ZeroPad2d(1), tnn.ZeroPad2d(1))
        check_equal(arr, jnn.ZeroPad2d((2,3,34,1)), tnn.ZeroPad2d((2,3,34,1)))

        # ***************************************************************
        # Test ReflectionPad2d Layer
        # ***************************************************************
        arr = np.random.randn(16,3,224,224)
        check_equal(arr, jnn.ReflectionPad2d(20), tnn.ReflectionPad2d(20))
        check_equal(arr, jnn.ReflectionPad2d((2,3,34,1)), tnn.ReflectionPad2d((2,3,34,1)))
        check_equal(arr, jnn.ReflectionPad2d((10,123,34,1)), tnn.ReflectionPad2d((10,123,34,1)))
        check_equal(arr, jnn.ReflectionPad2d((100)), tnn.ReflectionPad2d((100)))

        # ***************************************************************
        # Test function pad
        # ***************************************************************
        arr = np.random.randn(16,3,224,224)
        padding = (10,11,2,3)
        for mode in ['constant','replicate','reflect','circular']:
            j_data = jt.array(arr)
            t_data = torch.tensor(arr)
            t_output = tnn.functional.pad(t_data,padding,mode=mode).detach().numpy()
            j_output = jnn.pad(j_data,padding,mode).numpy()
            assert np.allclose(t_output,j_output)


if __name__ == "__main__":
    unittest.main()


# ***************************************************************
# Copyright (c) 2023 Jittor. All Rights Reserved. 
# Maintainers: Dun Liang <randonlang@gmail.com>. 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import jittor as jt
import numpy as np
import os

def expect_error(func):
    try:
        func()
    except Exception as e:
        return
    raise Exception("Expect an error, but nothing catched.")

class TestCore(unittest.TestCase):

    def test_number_of_hold_vars(self):
        assert jt.random([1,2,3]).peek() == "float32[1,2,3,]"
        assert jt.core.number_of_hold_vars() == 0
        x = jt.random([1,2,3])
        assert jt.core.number_of_hold_vars() == 1
        del x
        assert jt.core.number_of_hold_vars() == 0

    def test_fetch_sync(self):
        dtypes = ["float32", "float64"]
        for dtype in dtypes:
            x = jt.random([1,2,3], dtype)
            res = x.data
            assert res.dtype == dtype and res.shape == (1,2,3)

    def test_set_seed(self):
        a = jt.random([1,2,3]).data
        b = jt.random([1,2,3]).data
        assert str(a) != str(b)
        jt.set_seed(1)
        a = jt.random([1,2,3]).data
        jt.set_seed(1)
        b = jt.random([1,2,3]).data
        assert str(a) == str(b)
        
    def test_array_op(self):
        data = [
            np.array([1,2,3]),
            np.int32([1,2,3]),
            np.int64([1,2,3]),
            np.float32([1,2,3]),
            np.float64([1,2,3]),
        ]
        for a in data:
            assert sum(jt.array(a).data) == 6
        assert np.all(jt.array(np.int32([1,2,3])[::-1]).data == [3,2,1])
        assert jt.array(1).data.shape == (1,)
        
    def test_matmul_op(self):
        a = np.array([[1, 0], [0, 1]]).astype("float32")
        b = np.array([[4, 1], [2, 2]]).astype("float32")
        c = np.matmul(a, b)
        jtc = jt.matmul(jt.array(a), jt.array(b)).data
        assert np.allclose(jtc, c)

        a = np.random.random((128,3,10,20))
        b = np.random.random((20,30))
        c = np.matmul(a, b)
        jtc = jt.matmul(jt.array(a), jt.array(b)).data
        assert np.allclose(jtc, c)

        a = np.random.random((128,3,10,20))
        b = np.random.random((128,3,20,30))
        c = np.matmul(a, b)
        jtc = jt.matmul(jt.array(a), jt.array(b)).data
        assert np.allclose(jtc, c), np.abs(jtc-c).max()

    def test_var_holder(self):
        jt.clean()
        self.assertEqual(jt.number_of_lived_vars(), 0)
        expect_error(lambda: jt.matmul(1,1))
        expect_error(lambda: jt.matmul([1],[1]))
        expect_error(lambda: jt.matmul([[1]],[1]))
        self.assertEqual(jt.number_of_lived_vars(), 0)
        a = jt.matmul(jt.float32([[3]]), jt.float32([[4]])).data
        assert a.shape == (1,1) and a[0,0] == 12
        a = np.array([[1, 0], [0, 1]]).astype("float32")
        b = np.array([[4, 1], [2, 2]]).astype("float32")
        c = np.matmul(a, b)
        jtc = jt.matmul(jt.array(a), jt.array(b)).data
        assert np.all(jtc == c)

    def test_save_load_sub_module(self):
        class Net(jt.Module):
            def __init__(self):
                self.conv1 = jt.nn.Conv(3,3,3)
        net = Net()
        assert list(net.state_dict().keys()) == ['conv1.weight', 'conv1.bias']
        assert list(net.conv1.state_dict().keys()) == ['weight', 'bias']
        pkl_name = os.path.join(jt.flags.cache_path, "sub.pkl")
        net.conv1.save(pkl_name)
        net.conv1.load(pkl_name)

    def test_module(self):
        a = jt.Module()
        a.__setattr__("x", 1)
        assert a.__getattr__("x") == 1
        a.y = 2
        assert a.y == 2

    def test_modules(self):
        a = jt.Module()
        a.x = jt.Module()
        a.y = jt.Module()
        a.a = jt.array([1,2,3])
        a.b = jt.array([1,2,3])
        assert list(a._modules.keys()) == ["x", "y"]
        assert a._modules['x'] is a.x
        assert a._modules['y'] is a.y
        assert list(a._parameters.keys()) == ['a', 'b']
        assert a._parameters['a'] is a.a
        assert a._parameters['b'] is a.b

    def test_copy_memopt(self):
        # exe: post run
        # remove pending done
        # add hold pending done
        # pending release mem done
        a = jt.rand(10)
        b = a.copy().copy().copy()
        a.name("aa")
        b.name("bb")

        cnt = 0
        graphs = jt.dump_all_graphs()
        for x in graphs.nodes_info:
            if "Var" not in x: continue
            print(x)
            if ",aa," in x:
                assert ":2:i" in x, x
            elif ",bb," in x:
                assert ":1:i" in x
            else:
                assert ":1:i" in x

        b.sync()
        cnt = 0
        graphs = jt.dump_all_graphs()
        for x in graphs.nodes_info:
            # print(x)
            if "Var" in x and ",0)" in x:
                cnt += 1
        assert cnt == 2

    def test_fuse_memopt(self):
        def check():
            a = jt.rand(10)
            b = (a.copy().name("copy_out1") + 1).sqr() + a.copy().name("copy_out2")
            b.sync()
            for n in jt.dump_all_graphs().nodes_info:
                if "Var" not in n: continue
                # print(n)

                if "copy_out1" in n:
                    # copy out1 is not free
                    assert ",0)" not in n
                if "copy_out2" in n:
                    # copy out2 is freeed
                    assert ",0)" in n
            da = jt.grad(b, a)
            da.sync()
        check()
        jt.gc()
        assert jt.liveness_info()['lived_vars'] == 0

    def test_out_hint1(self):
        a = jt.rand(10)
        b = jt.rand(10)
        c = jt.ternary_out_hint((a<b).out_hint(), a, b).clone()
        c.sync()
        da, db = jt.grad(c, [a, b])
        jt.sync_all()
        for n in jt.dump_all_graphs().nodes_info:
            if "Var" in n and "bool" in n:
                print(n)
                assert ",0)" not in n
        jt.ternary_out_hint((a<b).out_hint(), a, 0).sync()

    def test_out_hint2(self):
        a = jt.rand(10)
        b = jt.rand(10)
        c = jt.ternary(a<b, a, b).clone()
        # c.sync()
        da, db = jt.grad(c, [a, b])
        jt.sync_all()
        for n in jt.dump_all_graphs().nodes_info:
            if "Var" in n and "bool" in n:
                print(n)
                assert ",0)" not in n

    def test_relu_memopt(self):
        x = a = jt.rand(10,10)
        for i in range(10):
            # a = jt.nn.relu(a)
            a = jt.ternary_out_hint((a>0.0).name("b"+str(i)), a, 0.0)
            a = jt.matmul(a.name("m1"),jt.rand(10,10).name("m2")).name("m3-"+str(i))
        da = jt.grad(a, x, True)
        # jt.clean_graph()
        da.sync()
        cnt1 = 0
        cnt2 = 0
        for n in jt.dump_all_graphs().nodes_info:
            if "Var" in n and ",0)" not in n:
                cnt1 +=1
                if "bool" in n:
                    cnt2 += 1
        print(cnt1, cnt2)
        assert cnt2 == 10
        assert cnt1 <= 33, cnt1

    def test_node_order(self):
        a = jt.nn.Sequential()
        for i in range(10):
            a.append(jt.nn.Linear(10,10, bias=False))
        sgd = jt.optim.SGD(a.parameters(), 0.1)
        jt.sync_all()
        with jt.log_capture_scope(log_silent=1,
                log_vprefix="exe=100") as logs:
            x = jt.rand(3,10)
            y = a(x)
            sgd.step(y*y)
            jt.sync_all()
        orders = []
        for l in logs:
            msg = l["msg"]
            if "Finished" in msg:
                # print(msg)
                if "weight" in msg:
                    assert msg.count("Var") >= 2
                    order = int(msg.split('fused ')[1].split("/")[0])
                    # print(order)
                    orders.append(order)
        assert len(orders) == 10, orders
        for i in range(10):
            assert orders[i] <= 14+i*3

    def test_bc_bug(self):
        a = jt.zeros((1,1))
        b = a * 0.5
        b.sync()
        da = jt.grad(b, a)
        da.sync()

    def test_attr_dict(self):
        a = jt.array([1,2,3])
        a.hahaha = 1
        assert a.hahaha == 1

    def test_swap(self):
        # TODO: add skip
        return
        jt.gc()
        jt.display_memory_info()
        with jt.flag_scope(cpu_mem_limit=50*1024*1024):
            np_arrays = []
            jt_arrays = []
            for i in range(100):
                jt_arrays.append(jt.randn(1*1024*1024//4))
                np_arrays.append(jt_arrays[-1].numpy())
                for j in range(i):
                    is_swapped = jt_arrays[j].location() == "disk"
                    assert is_swapped == (j<=i-50), (i,j,jt_arrays[j].debug_msg(), is_swapped,  j<=i-50, jt.display_memory_info())
            for i in range(len(np_arrays)):
                np.testing.assert_allclose(jt_arrays[i].numpy(), np_arrays[i])
            for i in range(len(np_arrays)):
                np.testing.assert_allclose(jt_arrays[i].numpy(), np_arrays[i])

    def test_swap_cuda(self):
        # TODO: add skip
        return
        jt.gc()
        jt.display_memory_info()
        if jt.has_cuda:
            np_arrays = []
            jt_arrays = []
            jt.gc()
            jt.display_memory_info()
            cpu_mem_limit=50*1024*1024
            device_mem_limit=20*1024*1024
            with jt.flag_scope(use_cuda=1, 
                cpu_mem_limit=cpu_mem_limit,
                device_mem_limit=device_mem_limit):
                for i in range(100):
                    jt_arrays.append(jt.random([1*1024*1024//4]))
                    np_arrays.append(jt_arrays[-1].numpy())
                    jt_arrays[-1].sync().name(str(i))
                    meminfo = jt.get_mem_info()
                    assert meminfo.total_cpu_used < cpu_mem_limit + 2*1024*1024
                    assert meminfo.total_cuda_used < device_mem_limit + 2*1024*1024
                for i in range(len(np_arrays)):
                    np.testing.assert_allclose(jt_arrays[i].numpy(), np_arrays[i])
                    meminfo = jt.get_mem_info()
                    assert meminfo.total_cpu_used < cpu_mem_limit + 2*1024*1024
                    assert meminfo.total_cuda_used < device_mem_limit + 2*1024*1024
                for i in range(len(np_arrays)):
                    np.testing.assert_allclose(jt_arrays[i].numpy(), np_arrays[i])
                    meminfo = jt.get_mem_info()
                    assert meminfo.total_cpu_used < cpu_mem_limit + 2*1024*1024
                    assert meminfo.total_cuda_used < device_mem_limit + 2*1024*1024



if __name__ == "__main__":
    unittest.main()


# ***************************************************************
# Copyright (c) 2023 Jittor. All Rights Reserved. 
# Maintainers: 
#     Wenyang Zhou <576825820@qq.com>. 
# 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import jittor as jt
import numpy as np
skip_this_test = False

@unittest.skipIf(skip_this_test, "No Torch found")
class TestSetitem(unittest.TestCase):
    def test_setitem_(self):
        arr0 = jt.random((4,2,2))
        data0 = jt.ones((2,2))
        arr0[1] = data0
        arr0.sync()
        data0.data[0,0] = 0
        assert arr0[1,0,0] == 0

        arr00 = jt.random((4,2,2))
        data00 = jt.ones((2,2))
        # share memory will fail if d has an edge to other nodes.
        tmp = data00 + 1
        arr00[1] = data00
        arr00.sync()
        data00.data[0,0] = 0
        assert arr00[1,0,0] == 0

        arr1 = jt.random((4,2,2))
        data1 = jt.zeros((2,2))
        arr1[3,:,:] = data1
        arr1.sync()
        data1.data[0,0] = 1
        assert arr1[3,0,0] == 1

        arr21 = jt.ones((2,2))
        arr22 = jt.ones((2,2)) * 2
        arr2 = jt.concat([arr21, arr22], dim=0)
        arr2.sync()
        arr21.data[0,0] = 3
        arr22.data[0,0] = 4
        assert arr2[0,0] == 3
        assert arr2[2,0] == 4

    def test_getitem(self):
        # test for different slice type
        arr0 = jt.random((4,3))
        arr0_res = arr0[2,:]
        arr0_res.data[1] = 1
        assert arr0[2,1] == 1

        arr1 = jt.array([1,2,3,4])
        arr1_res = arr1[None]
        arr1_res.data[0,2] = -1
        assert arr1[2] == -1

        arr2 = jt.array([1,2,3,4])
        arr2_res = arr2[...]
        arr2_res.data[2] = -1
        assert arr2[2] == -1

        arr3 = jt.array([1,2,3,4])
        arr3_res = arr3[3]
        arr3_res.data[0] = -1
        assert arr3[3] == -1

        arr4 = jt.random((4,2,3,3))
        arr4_res = arr4[...,:,:]
        arr4_res.data[0,0,1,1] = 1
        assert arr4[0,0,1,1] == 1

        arr4 = jt.random((4,2,3,3))
        arr4_res = arr4[...,:,:2]
        arr4_res.data[0,0,1,1] = 1
        assert arr4[0,0,1,1] != 1

        arr4 = jt.random((3,3))
        arr4_res = arr4[...,:,:2]
        arr4_res.data[1,1] = 1
        assert arr4[1,1] != 1

        arr5 = jt.random((4,2,3,3))
        arr5_res = arr5[1:3,:,:,:]
        arr5_res.data[1,0,1,1] = 1
        assert arr5[2,0,1,1] == 1

        arr6 = jt.random((4,2,3,3))
        arr6_res = arr6[1]
        arr6_res.data[0,1,1] = 1
        assert arr6[1,0,1,1] == 1

        # test for different data type (float32/float64/bool/int8/int32)
        arr_float32 = jt.random((4,2,3))
        arr_float32_res = arr_float32[1:3,:,:]
        arr_float32_res.data[0,0,0] = 1
        assert arr_float32[1,0,0] == 1
        arr_float32_res.data[1,1,2] = 1
        assert arr_float32[2,1,2] == 1
        arr_float32[1,0,0] = 0
        # getitem and setitem do not conflict 
        assert arr_float32_res[0,0,0] == 1

        arr_bool = jt.bool(np.ones((4,2,3)))
        arr_bool_res = arr_bool[1:3,:,:]
        arr_bool_res.data[0,0,0] = False
        assert arr_bool[1,0,0] == False
        arr_bool_res.data[0,0,1] = False
        assert arr_bool[1,0,1] == False

        arr_float64 = jt.random((4,2,3), dtype='float64')
        arr_float64_res = arr_float64[1:3,:,:]
        arr_float64_res.data[0,0,0] = 1
        assert arr_float64[1,0,0] == 1
        arr_float64_res.data[1,1,2] = 1
        assert arr_float64[2,1,2] == 1

        arr_int32 = jt.ones((4,2,3), dtype='int32')
        arr_int32_res = arr_int32[1:3,:,:]
        arr_int32_res.data[0,0,0] = 0
        assert arr_int32[1,0,0] == 0
        arr_int32_res.data[1,1,2] = 0
        assert arr_int32[2,1,2] == 0

    def test_setitem_inplace_case1(self):
        # test type case
        a = jt.zeros((3,))
        a[1] = 123
        assert a.data[1] == 123

    def test_setitem_inplace_case2(self):
        # test un-continuous first dim
        a = jt.zeros((3,))
        a[0::2] = jt.ones((2,))
        assert a.data[2] == 1

    def test_setitem_inplace_case3(self):
        # test broadcast
        a = jt.zeros((3,))
        a[0:] = 1.0
        assert a.data[2] == 1

    @unittest.skipIf(not jt.compiler.has_cuda, "No CUDA found")
    @jt.flag_scope(use_cuda=1)
    def test_getitem_inplace_array(self):
        a = jt.array([[1,2],[3,4]])
        assert (a[0].numpy() == [1,2]).all(), a[0].numpy()
        assert (a[1].numpy() == [3,4]).all(), a[1].numpy()

    @unittest.skipIf(not jt.compiler.has_cuda, "No CUDA found")
    @jt.flag_scope(use_cuda=1)
    def test_setitem_inplace_array(self):
        a = jt.array([[1,2],[3,4]])
        a[0,0] = -1
        a[1,1] = -2
        assert (a[0].numpy() == [-1,2]).all(), a[0].numpy()
        assert (a[1].numpy() == [3,-2]).all(), a[1].numpy()

    def test_scatter(self):
        src = jt.arange(1, 11).reshape((2, 5))
        index = jt.array([[0, 1, 2, 0]])
        x = jt.zeros((3, 5), dtype=src.dtype).scatter_(0, index, src)
        assert (x.data == 
            [[1, 0, 0, 4, 0],
            [0, 2, 0, 0, 0],
            [0, 0, 3, 0, 0]]).all()
        index = jt.array([[0, 1, 2], [0, 1, 4]])
        x = jt.zeros((3, 5), dtype=src.dtype).scatter_(1, index, src)
        assert (x.data ==
            [[1, 2, 3, 0, 0],
            [6, 7, 0, 0, 8],
            [0, 0, 0, 0, 0]]).all()
        x = jt.full((2, 4), 2.).scatter_(1, jt.array([[2], [3]]),
               jt.array(1.23), reduce='multiply')
        assert np.allclose(x.data, 
            [[2.0000, 2.0000, 2.4600, 2.0000],
            [2.0000, 2.0000, 2.0000, 2.4600]]), x
        x = jt.full((2, 4), 2.).scatter_(1, jt.array([[2], [3]]),
               jt.array(1.23), reduce='add')
        assert np.allclose(x.data,
            [[2.0000, 2.0000, 3.2300, 2.0000],
            [2.0000, 2.0000, 2.0000, 3.2300]])

    def test_gather(self):
        t = jt.array([[1, 2], [3, 4]])
        data = t.gather(1, jt.array([[0, 0], [1, 0]])).data
        assert (data == [[ 1,  1], [ 4,  3]]).all()
        data = t.gather(0, jt.array([[0, 0], [1, 0]])).data
        assert (data == [[ 1,  2], [ 3,  2]]).all()

    @unittest.skipIf(not jt.compiler.has_cuda, "No CUDA found")
    @jt.flag_scope(use_cuda=1)
    def test_scatter_cuda(self):
        self.test_scatter()

    @unittest.skipIf(not jt.compiler.has_cuda, "No CUDA found")
    @jt.flag_scope(use_cuda=1)
    def test_gather_cuda(self):
        self.test_gather()

    def test_setitem_bool(self):
        a = jt.array([1,2,3,4])
        b = jt.array([True,False,True,False])
        a[b] = jt.array([-1,-2])
        assert (a.data == [-1,2,-2,4]).all()

    def test_setitem_bool2(self):
        a = jt.array([1,2,3,4])
        b = jt.array([True,False,True,False])
        a[b] = jt.array([-1])
        assert (a.data == [-1,2,-1,4]).all(), a
        a = jt.array([1,2,3,4])
        b = jt.array([True,False,True,False])
        a[b] = -1
        assert (a.data == [-1,2,-1,4]).all(), a
        
    def test_slice_none(self):
        a = jt.array([1,2])
        assert a[None,:,None,None,...,None].shape == (1,2,1,1,1)

    def test_roll(self):
        x = jt.array([1, 2, 3, 4, 5, 6, 7, 8]).view(4, 2)
        y = x.roll(1, 0)
        assert (y.numpy() == [[7,8],[1,2],[3,4],[5,6]]).all(), y
        y = x.roll(-1, 0)
        assert (y.numpy() == [[3,4],[5,6],[7,8],[1,2]]).all()
        y = x.roll(shifts=(2, 1), dims=(0, 1))
        assert (y.numpy() == [[6,5],[8,7],[2,1],[4,3]]).all()

    def test_ellipsis_with_none(self):
        a = jt.arange(2*4*4).reshape(2,4,4)
        b = a[...,:,None,:2]
        assert b.shape == [2,4,1,2]
        np.testing.assert_allclose(b.data, a.data[...,:,None,:2])

    def test_flip_grad(self):
        a = jt.rand(10)
        b = a[::-1]
        c = b[::-1]
        d = c.sum()
        jt.grad(d, [a])

    def test_concat2(self):
        a = jt.rand(10)
        b = jt.rand(11)
        c = jt.rand(12)
        def cc():
            x = jt.concat([b.copy(), c.copy()])
            d = jt.concat([a.copy(), x])
            return d.copy().copy().copy().copy().copy().copy()\
                .copy().copy() + x.sum()*0.0
        d = cc()
        np.testing.assert_allclose(d.data,
            np.concatenate([a.data,b.data,c.data]))

    def test_concat3(self):
        # a = jt.rand(10)
        b = jt.rand(11)
        c = jt.rand(12)
        def cc():
            x = jt.concat([b.copy(), c.copy()])
            d = jt.concat([x])
            return d.copy().copy().copy().copy().copy().copy()\
                .copy().copy() + x.sum()*0.0
        d = cc()
        np.testing.assert_allclose(d.data,
            np.concatenate([b.data,c.data]))
        

    def test_concat4(self):
        # a = jt.rand(10)
        b = jt.rand(11)
        c = jt.rand(12)
        def cc():
            x = jt.concat([b.copy(), c.copy()])
            d = jt.concat([x])
            return d
        d = cc()
        np.testing.assert_allclose(d.data,
            np.concatenate([b.data,c.data]))
        
    def test_concat_random(self):
        def check(backward=False):
            n1, n2, n3 = 1000, 20, 10
            # n1, n2, n3 = 3, 2, 3
            import random
            data = []
            back = []
            for i in range(n1):
                if len(data) > n2:
                    v = random.randint(0,len(data)-1)
                    # print("del", v)
                    del data[v]
                x1 = random.randint(0,9)
                # print(i, x1)
                if len(data) == 0:
                    # a = jt.random((random.randint(10,20),))
                    a = jt.array(np.random.rand(random.randint(n3,n3*2)))
                    data.append(a)
                if x1 == 0:
                    a = data[random.randint(0,len(data)-1)]
                    a = a.copy()
                    data.append(a)
                elif x1 == 1:
                    a = data[random.randint(0,len(data)-1)]
                    a = a.clone()
                    data.append(a)
                elif x1 == 2:
                    a = data[random.randint(0,len(data)-1)]
                    b = np.random.permutation(np.arange(a.numel()))
                    # print("permutation", b)
                    a = a[b]
                    data.append(a)
                elif x1 == 3:
                    a = data[random.randint(0,len(data)-1)]
                    a = a[:100]
                    # print(a.shape)
                    data.append(a)
                elif x1 == 4:
                    # a = jt.random((random.randint(10,20),))
                    a = jt.array(np.random.rand(random.randint(n3,n3*2)))
                    if backward and random.randint(0,1):
                        back.append(a)
                    data.append(a)
                elif x1 == 5:
                    v = random.randint(0,len(data)-1)
                    a = data[v]
                    # print("split", v, a.shape)
                    arr = a.split(n3-1)
                    data += arr
                else:
                    if not len(data): continue
                    n = random.randint(1,3)
                    a = [ data[random.randint(0,len(data)-1)] for i in range(n) ]
                    a = jt.concat(a)
                    if a.numel() > 1000:
                        b = np.random.permutation(np.arange(a.numel()))
                        a = a[b][:100]
                    data.append(a)
            ret = jt.concat(data)
            if backward and len(back):
                grads = jt.grad(jt.rand_like(ret)*ret, back)
                return jt.concat(grads).numpy()
            return ret.numpy()

        for s in range(100):
            print("check", s)
            for check_grad in [True, False]:
                jt.set_global_seed(s)
                data = check(check_grad)
                jt.gc()
                jt.set_global_seed(s)
                with jt.flag_scope(gopt_disable=1):
                    data2 = check(check_grad)
                jt.gc()
                np.testing.assert_allclose(data, data2, atol=1e-5, rtol=1e-5)

    def test_concat_grad(self):
        n = 30000
        m = 100
        arr = []
        for i in range(n):
            arr.append(jt.random((m,)))
        x = jt.concat(arr)
        y = jt.rand_like(x)
        grads = jt.grad(x*y, arr)
        for i in range(n):
            np.testing.assert_allclose(grads[i].numpy(), y[i*m:(i+1)*m].numpy())

    def test_split_grad(self):
        n = 30000
        m = 100
        x = jt.random((n*m,))
        arr = x.split(m)
        yy = [ jt.rand(m) for i in range(n) ]
        arr2 = [ y*yy[i] for i,y in enumerate(arr) ]
        g = jt.grad(jt.concat(arr2), x)
        for i in range(n):
            np.testing.assert_allclose(g.data[i*m:(i+1)*m], yy[i].data)

    def test_dfs_memopt(self):
        with jt.flag_scope(profile_memory_enable=1):
            n = 1024
            b = []
            for i in range(n):
                a = jt.rand(n).copy().copy()
                a = a.sum()
                # a.sync()
                b.append(a)
            jt.sync_all()
            jt.get_max_memory_treemap()


    def test_setitem_bc(self):
        a = jt.random([10,11,12])
        b = a[jt.arange(3)[:,None],
            jt.arange(4)[None,:]]
        b.sync()
        assert (a[:3, :4] == b).all()
        
        a = jt.random([10,11,12])
        b = a[jt.arange(3)[:,None],
            jt.arange(4)[None,:],
            jt.arange(4)[None,:]]
        nb = a.data[np.arange(3)[:,None],
            np.arange(4)[None,:],
            np.arange(4)[None,:]]
        np.testing.assert_allclose(nb, b.data)
        
        a = jt.random([10,11,12])
        b = a[jt.arange(3)[::-1,None],
            jt.arange(4)[None,:],
            jt.arange(4)[None,:]]
        nb = a.data[np.arange(3)[::-1,None],
            np.arange(4)[None,:],
            np.arange(4)[None,:]]
        np.testing.assert_allclose(nb, b.data)
        
        a = jt.random([10,11,12])
        b = a[jt.arange(3)[::-1,None],
            jt.arange(4)[None,:],
            jt.arange(4)[None,::-1]]
        nb = a.data[np.arange(3)[::-1,None],
            np.arange(4)[None,:],
            np.arange(4)[None,::-1]]
        np.testing.assert_allclose(nb, b.data)

    def test_cuda_slice_migrate_bug(self):
        a = jt.array([1,2,3,4,5])
        jt.sync_all()
        if not jt.has_cuda: return
        with jt.flag_scope(use_cuda=1):
            b = a[0]
            b.sync(True)
            assert b.item() == 1

    def test_cascade_setitem(self):
        a = jt.zeros(3,3,3,3)
        a[1][2][0][0] = 1
        assert a[1,2,0,0] == 1
        # TODO: convert a[x] = a[x] + b -> a[x] += b
        a[1][2][0][0] += 1
        assert a[1,2,0,0] == 2

if __name__ == "__main__":
    unittest.main()



# ***************************************************************
# Copyright (c) 2023 Jittor. All Rights Reserved. 
# Maintainers: 
#     Wenyang Zhou <576825820@qq.com>
#     Dun Liang <randonlang@gmail.com>. 
# 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import jittor as jt
import numpy as np
import jittor.nn as jnn

skip_this_test = False

try:
    jt.dirty_fix_pytorch_runtime_error()
    import torch
    import torch.nn as tnn
except:
    torch = None
    tnn = None
    skip_this_test = True

def check_equal_with_istrain(arr, j_layer, p_layer, is_train=True, has_running=True, threshold=1e-5):
    jittor_arr = jt.array(arr)
    pytorch_arr = torch.Tensor(arr)
    if has_running:
        if is_train:
            assert np.allclose(p_layer.running_mean.detach().numpy(), j_layer.running_mean.numpy(), threshold)
        else:
            assert np.allclose(p_layer.layer.running_mean.detach().numpy(), j_layer.running_mean.numpy(), threshold)
    jittor_result = j_layer(jittor_arr)
    pytorch_result = p_layer(pytorch_arr)
    if has_running:
        if is_train:
            assert np.allclose(p_layer.running_mean.detach().numpy(), j_layer.running_mean.numpy(), threshold)
        else:
            assert np.allclose(p_layer.layer.running_mean.detach().numpy(), j_layer.running_mean.numpy(), threshold)
    assert np.allclose(pytorch_result.detach().numpy(), jittor_result.numpy(), 1e-2, threshold), \
        ( np.abs(pytorch_result.detach().numpy() - jittor_result.numpy()).max() )

def check_equal_without_istrain(arr, j_layer, p_layer, threshold=1e-5):
    jittor_arr = jt.array(arr)
    pytorch_arr = torch.Tensor(arr)
    jittor_result = j_layer(jittor_arr)
    pytorch_result = p_layer(pytorch_arr)
    assert np.allclose(pytorch_result.detach().numpy(), jittor_result.numpy(), threshold)

@unittest.skipIf(skip_this_test, "No Torch found")
class TestBatchNorm(unittest.TestCase):
    @jt.flag_scope(auto_convert_64_to_32=0)
    def test_batchnorm(self):
        # ***************************************************************
        # Test BatchNorm Layer
        # ***************************************************************
        arr = np.random.randn(16,10,224,224)
        check_equal_with_istrain(arr, jnn.BatchNorm(10, is_train=True), tnn.BatchNorm2d(10))

        class Model(tnn.Module):
            def __init__(self):
                super(Model, self).__init__()
                self.layer = tnn.BatchNorm2d(10)
            def forward(self, x):
                return self.layer(x)
        model = Model()
        model.eval()
        check_equal_with_istrain(arr, jnn.BatchNorm(10, is_train=False), model, False)

        # ***************************************************************
        # Test InstanceNorm2d Layer
        # ***************************************************************
        arr = np.random.randn(16,10,224,224)
        check_equal_without_istrain(arr, jnn.InstanceNorm2d(10, is_train=True), tnn.InstanceNorm2d(10))

        class Model(tnn.Module):
            def __init__(self):
                super(Model, self).__init__()
                self.layer = tnn.InstanceNorm2d(10)
            def forward(self, x):
                return self.layer(x)
        model = Model()
        model.eval()
        check_equal_without_istrain(arr, jnn.InstanceNorm2d(10, is_train=False), model)

        # ***************************************************************
        # Test BatchNorm1d Layer
        # ***************************************************************
        arr = np.random.randn(16,10)
        check_equal_with_istrain(arr, jnn.BatchNorm1d(10, is_train=True), tnn.BatchNorm1d(10), 1e-3)

        class Model(tnn.Module):
            def __init__(self):
                super(Model, self).__init__()
                self.layer = tnn.BatchNorm1d(10)
            def forward(self, x):
                return self.layer(x)
        model = Model()
        model.eval()
        check_equal_with_istrain(arr, jnn.BatchNorm1d(10, is_train=False), model, False)

        # ***************************************************************
        # Test GroupNorm Layer
        # ***************************************************************
        arr = np.random.randn(16,10,224,224)

        class Model(tnn.Module):
            def __init__(self):
                super(Model, self).__init__()
                self.layer = tnn.GroupNorm(2, 10)
            def forward(self, x):
                return self.layer(x)
        model = Model()
        model.eval()
        check_equal_with_istrain(arr, jnn.GroupNorm(2, 10, is_train=False), model, False, False)

        # ***************************************************************
        # Test LayerNorm Layer
        # ***************************************************************
        arr = np.random.randn(16,10,224,224)

        class Model(tnn.Module):
            def __init__(self):
                super(Model, self).__init__()
                self.layer = tnn.LayerNorm(224)
            def forward(self, x):
                return self.layer(x)
        model = Model()
        model.eval()
        check_equal_with_istrain(arr, jnn.LayerNorm(224), model, False, False)

if __name__ == "__main__":
    unittest.main()


# ***************************************************************
# Copyright (c) 2023 Jittor. All Rights Reserved. 
# Maintainers: 
#     Wenyang Zhou <576825820@qq.com>
#     Dun Liang <randonlang@gmail.com>. 
# 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import os, sys
import jittor as jt
import numpy as np
from jittor.test.test_mpi import run_mpi_test
mpi = jt.compile_extern.mpi

from jittor.dataset.mnist import MNIST

def val1():
    dataloader = MNIST(train=False).set_attrs(batch_size=16)
    for i, (imgs, labels) in enumerate(dataloader):
        assert(imgs.shape[0]==8)
        if i == 5:
            break

@jt.single_process_scope(rank=0)
def val2():
    dataloader = MNIST(train=False).set_attrs(batch_size=16)
    for i, (imgs, labels) in enumerate(dataloader):
        assert(imgs.shape[0]==16)
        if i == 5:
            break

@unittest.skipIf(not jt.in_mpi, "no inside mpirun")
class TestSingleProcessScope(unittest.TestCase):
    def test_single_process_scope(self):
        val1()
        val2()

@unittest.skipIf(not jt.compile_extern.has_mpi, "no mpi found")
class TestSingleProcessScopeEntry(unittest.TestCase):
    def test_entry(self):
        run_mpi_test(2, "test_single_process_scope")

if __name__ == "__main__":
    unittest.main()


from copy import deepcopy
from pathlib import Path
import jittor as jt
import jittor.nn as nn
import numpy as np
import os

split_size = 1000000

conv_opt = int(os.environ.get("conv_opt", "0"))

if conv_opt:
    Conv1d_sp = nn.Conv1d_sp
else:
    Conv1d_sp = nn.Conv1d


def MLP(channels: list, do_bn=True):
    """ Multi-layer perceptron """
    n = len(channels)
    layers = []
    for i in range(1, n):
        layers.append(Conv1d_sp(channels[i - 1], channels[i], kernel_size=1, bias=True))
        if i < (n - 1):
            if do_bn:
                layers.append(nn.BatchNorm(channels[i]))
                # layers.append(nn.InstanceNorm1d(channels[i]))
                # layers.append(nn.LayerNorm(channels[i]))
            layers.append(nn.ReLU())
    return nn.Sequential(*layers)


def normalize_keypoints(kpts, image_shape):
    size = image_shape.flip(1)  # shape=(b,2) ;h w -> w, h
    center = size / 2
    scaling = size.float32().max(1, keepdims=True) * 0.7
    return (kpts - center[:, None, :]) / scaling[:, None, :]


class KeypointEncoder(nn.Module):
    """ Joint encoding of visual appearance and location using MLPs"""
    def __init__(self, feature_dim, layers, keypoint_position_dim=2):
        super().__init__()
        # self.keypoint_position_dim = keypoint_position_dim
        self.encoder = MLP([keypoint_position_dim + 1] + layers + [feature_dim])
        nn.init.constant_(self.encoder[-1].bias, 0.0)

    def execute(self, kpts, scores):
        inputs = jt.concat([kpts.t(), scores.unsqueeze(1)], dim=1)
        return self.encoder(inputs)

cnt = 0

def attention(query, key, value):
    global cnt
    cnt += 1
    b, d, h, n = query.shape
    # print("attention", b,d,h,n, cnt)
    dim_factor = (1.0 / d)**0.5
    query = query.transpose(0, 2, 3, 1).reshape(b * h, -1, d) * dim_factor
    key = key.transpose(0, 2, 1, 3).reshape(b * h, d, -1)
    value = value.transpose(0, 2, 3, 1).reshape(b * h, -1, d)
    # print("attention", query.shape, key.shape, value.shape)

    data = []
    for i in range(0, query.shape[0], split_size):
        end = min(i + split_size, query.shape[0])
        tmp1 = nn.bmm(query[i:end], key[i:end])
        tmp2 = nn.softmax(tmp1, dim=-1)
        tmp3 = nn.bmm(tmp2, value[i:end])
        tmp3.sync()
        data.append(tmp3)
    tmp3 = jt.concat(data)
    
    # for i in range(0, query.shape[0], split_size):
    #     end = min(i + split_size, query.shape[0])
    #     tmp1 = nn.bmm(query[:,i:end], key[:,i:end])
    #     tmp2 = nn.softmax(tmp1, dim=-1)
    #     tmp3 = nn.bmm(tmp2, value[:,i:end])
    #     tmp3.sync()
    #     data.append(tmp3)
    # tmp3 = jt.concat(data, dim=1)

    # tmp1 = nn.bmm(query, key)
    # print(tmp1.shape)
    # tmp2 = nn.softmax(tmp1, dim=-1)
    # print(tmp2.shape)
    # tmp3 = nn.bmm(tmp2, value)
    # print(tmp3.shape)
    return tmp3.reshape(b, h, -1, d).transpose(0, 3, 1, 2)
    return nn.bmm(nn.softmax(nn.bmm(query, key), dim=-1), value).reshape(b, h, -1, d).transpose(0, 3, 1, 2)


class MultiHeadedAttention(nn.Module):
    """ Multi-head attention to increase model expressivitiy """
    def __init__(self, num_heads: int, d_model: int):
        super().__init__()
        assert d_model % num_heads == 0
        self.dim = d_model // num_heads
        self.num_heads = num_heads
        self.merge = Conv1d_sp(d_model, d_model, kernel_size=1)
        self.proj = nn.ModuleList([deepcopy(self.merge) for _ in range(3)])

    def execute(self, query, key, value):
        batch_dim = query.size(0)
        query, key, value = [l(x).reshape(batch_dim, self.dim, self.num_heads, -1) for l, x in zip(self.proj, (query, key, value))]
        x = attention(query, key, value)
        # x = attention_chunk(query, key, value)
        return self.merge(x.reshape(batch_dim, self.dim * self.num_heads, -1))


class AttentionalPropagation(nn.Module):
    def __init__(self, feature_dim: int, num_heads: int):
        super().__init__()
        self.attn = MultiHeadedAttention(num_heads, feature_dim)
        self.mlp = MLP([feature_dim * 2, feature_dim * 2, feature_dim])
        nn.init.constant_(self.mlp[-1].bias, 0.0)

    def execute(self, x, source):
        message = self.attn(x, source, source)
        return self.mlp(jt.concat([x, message], dim=1))


class AttentionalGNN(nn.Module):
    def __init__(self, feature_dim: int, layer_names: list):
        super().__init__()
        self.layers = nn.ModuleList([AttentionalPropagation(feature_dim, 4) for _ in range(len(layer_names))])
        self.is_cross = [x == 'cross' for x in layer_names]

    def execute(self, desc0, desc1):
        for layer, is_cross in zip(self.layers, self.is_cross):
            layer.attn.prob = []
            if is_cross:
                src0, src1 = desc1, desc0
            else:  # if name == 'self':
                src0, src1 = desc0, desc1
            # delta0, delta1 = layer(desc0, src0), layer(desc1, src1)

            delta0 = layer(desc0, src0)
            # print(delta0.numel()*4)
            # breakpoint()
            jt.sync_all()
            delta1 = layer(desc1, src1)
            jt.sync_all()
            desc0, desc1 = (desc0 + delta0), (desc1 + delta1)
            jt.sync_all()
        return desc0, desc1


def log_sinkhorn_iterations(Z, log_mu, log_nu, iters: int):
    """ Perform Sinkhorn Normalization in Log-space for stability"""
    u, v = jt.zeros_like(log_mu), jt.zeros_like(log_nu)
    for _ in range(iters):
        u = log_mu - (Z + v.unsqueeze(1)).exp().sum(dim=2).log()
        v = log_nu - (Z + u.unsqueeze(2)).exp().sum(dim=1).log()
    return Z + u.unsqueeze(2) + v.unsqueeze(1)


def log_optimal_transport(scores, alpha, iters: int):
    """ Perform Differentiable Optimal Transport in Log-space for stability"""
    b, m, n = scores.shape
    ms, ns = jt.float(m, requires_grad=False), jt.float(n, requires_grad=False)

    bins0 = alpha.broadcast([b, m, 1])
    bins1 = alpha.broadcast([b, 1, n])
    alpha = alpha.broadcast([b, 1, 1])

    couplings = jt.concat([jt.concat([scores, bins0], -1), jt.concat([bins1, alpha], -1)], 1)

    norm = -(ms + ns).log()
    log_mu = jt.concat([norm.broadcast([m]), ns.log() + norm])
    log_nu = jt.concat([norm.broadcast([n]), ms.log() + norm])
    log_mu, log_nu = log_mu[None].broadcast([b, m + 1]), log_nu[None].broadcast([b, n + 1])

    Z = log_sinkhorn_iterations(couplings, log_mu, log_nu, iters)
    Z = Z - norm  # multiply probabilities by M+N
    return Z


def arange_like(x, dim: int):
    return jt.ones(x.shape[dim], dtype=x.dtype)[None].cumsum()[0] - 1  # traceable in 1.1


default_config = {
    'descriptor_dim': 256,  # SuperPoint
    'weights': 'indoor',
    'keypoint_encoder': [32, 64, 128, 256],  # SuperPoint
    'GNN_layers': ['self', 'cross'] * 9,
    'sinkhorn_iterations': 100,
    'match_threshold': 0.2,
}


def get_weighted_loss_batch(scores, all_matches):
    matches0, matches1 = all_matches.chunk(chunks=2, dim=2)
    batchIdx = jt.arange(all_matches.shape[0]).unsqueeze(1).repeat(1, all_matches.shape[1])
    batchIdx, matches0, matches1 = batchIdx.view(-1), matches0.view(-1), matches1.view(-1)
    valid_index0, valid_index1 = matches0 >= 0, matches1 >= 0
    valid_match = jt.logical_and(valid_index0, valid_index1)
    valid_unmatch = jt.logical_xor(valid_index0, valid_index1)
    num_match = valid_match.sum().maximum(1e-9)
    num_unmatch = valid_unmatch.sum().maximum(1e-9)



    score_ = scores[batchIdx, matches0, matches1]
    score_match_ = (score_*valid_match).float32().sum() / num_match
    score_umatch_ = (score_*valid_unmatch).float32().sum() / num_unmatch
    return -(num_unmatch * score_match_ + num_match * score_umatch_) / (num_match + num_unmatch)
    # print(score_umatch_, score_match_)
    # return -(score_match + score_umatch) / (num_match + num_unmatch)

    score_match = scores[(batchIdx[valid_match], matches0[valid_match], matches1[valid_match])].float32().mean() if num_match > 0 else 0
    score_umatch = scores[(batchIdx[valid_unmatch], matches0[valid_unmatch], matches1[valid_unmatch])].float32().mean() if num_unmatch > 0 else 0
    # print(score_match, score_umatch)
    return -(num_unmatch * score_match + num_match * score_umatch) / (num_match + num_unmatch)


def add_dustbin(scores, alpha):
    b, m, n = scores.shape
    bins0 = jt.broadcast(alpha, (b, m, 1))
    bins1 = jt.broadcast(alpha, (b, 1, n))
    alpha = jt.broadcast(alpha, (b, 1, 1))
    couplings = jt.concat([jt.concat([scores, bins0], -1), jt.concat([bins1, alpha], -1)], 1)
    return couplings


class SuperGlue(nn.Module):
    def __init__(self, config):
        super().__init__()
        config = {**default_config, **config}
        self.descriptor_dim = config['descriptor_dim']
        self.keypoint_encoder = config['keypoint_encoder']
        self.GNN_layers = config['GNN_layers']
        self.sinkhorn_iterations = config['sinkhorn_iterations']
        self.match_threshold = config['match_threshold']
        self.keypoint_position_dim = config['keypoint_position_dim']
        self.use_dual_softmax = config['use_dual_softmax']
        self.scale = jt.float(self.descriptor_dim**-0.5).stop_grad()
        # self.scale.requires_grad = False

        # self.des_extend = MLP([128, 256])

        self.kenc = KeypointEncoder(self.descriptor_dim, self.keypoint_encoder, keypoint_position_dim=self.keypoint_position_dim)

        self.gnn = AttentionalGNN(self.descriptor_dim, self.GNN_layers)

        self.final_proj = Conv1d_sp(self.descriptor_dim, self.descriptor_dim, kernel_size=1, bias=True)

        self.bin_score = jt.float(1.0)

    def execute(self, data):
        """Run SuperGlue on a pair of keypoints and descriptors"""

        kpts0, kpts1 = data['keypoints0'], data['keypoints1']
        desc0, desc1 = data['descriptors0'], data['descriptors1']
        all_matches = data['all_matches']
        # match_num = data['match_num']

        if kpts0.shape[1] == 0 or kpts1.shape[1] == 0 or all_matches.shape[1] == 0:  # no keypoints or no matches/unmatches
            shape0, shape1 = kpts0.shape[:-1], kpts1.shape[:-1]
            return {
                'matches0': jt.ones(shape0, dtype=jt.int),
                'matches1': jt.ones(shape1, dtype=jt.int),
                'matching_scores0': jt.zeros(shape0, dtype=jt.float),
                'matching_scores1': jt.zeros(shape1, dtype=jt.float),
                'skip_train': True
            }

        # Keypoint normalization.
        kpts0 = normalize_keypoints(kpts0, data['shape0'])
        kpts1 = normalize_keypoints(kpts1, data['shape1'])

        # Keypoint MLP encoder.
        # desc0 = self.des_extend(desc0) + self.kenc(kpts0, data['scores0'])
        # desc1 = self.des_extend(desc1) + self.kenc(kpts1, data['scores1'])
        desc0 = desc0 + self.kenc(kpts0, data['scores0'])
        desc1 = desc1 + self.kenc(kpts1, data['scores1'])

        # Multi-layer Transformer network.
        desc0, desc1 = self.gnn(desc0, desc1)

        # Final MLP projection.
        desc0, desc1 = self.final_proj(desc0), self.final_proj(desc1)
        desc0_t = desc0.t()
        losses = []

        for i in range(0, desc1.shape[0], split_size):
            end = min(desc1.shape[0], i + split_size)

            # Compute matching descriptor distance.
            scores = nn.bmm(desc0_t[i:end], desc1[i:end]) * self.scale  # 457.76 MB
            scores.sync()

            # Run the optimal transport.
            if self.use_dual_softmax:
                scores = add_dustbin(scores, self.bin_score)  # 458.68 MB
                scores.sync()
                dual_softmax0, dual_softmax1 = nn.log_softmax(scores, 1), nn.log_softmax(scores, 2)
                scores = dual_softmax0 + dual_softmax1  # 458.22 MB
                scores.sync()
            else:
                scores = log_optimal_transport(scores, self.bin_score, iters=self.config['sinkhorn_iterations'])

            # loss = torch.stack([get_match_score(scores[b], all_matches[b]) for b in range(all_matches.shape[0])])

            loss = get_weighted_loss_batch(scores, all_matches[i:end])
            loss.sync()
            losses.append(loss)
        loss = jt.concat(losses)
        '''
        # Compute matching descriptor distance.
        scores = nn.bmm(desc0.t(), desc1) * self.scale # 457.76 MB
        scores.sync()

        # Run the optimal transport.
        if self.use_dual_softmax:
            scores = add_dustbin(scores, self.bin_score) # 458.68 MB
            scores.sync()
            dual_softmax0, dual_softmax1 = nn.log_softmax(scores, 1), nn.log_softmax(scores, 2)
            scores = dual_softmax0 + dual_softmax1 # 458.22 MB
            scores.sync()
        else:
            scores = log_optimal_transport(scores, self.bin_score, iters=self.config['sinkhorn_iterations'])

        # loss = torch.stack([get_match_score(scores[b], all_matches[b]) for b in range(all_matches.shape[0])])

        loss = get_weighted_loss_batch(scores, all_matches)
        # print(scores.shape, all_matches.shape, loss.shape)
        '''

        # matches0, matches1 = all_matches.chunk(chunks=2, dim=2)
        # batchIdx = jt.arange(0, b).unsqueeze(1).repeat(1, num)
        # batchIdx, matches0, matches1 = batchIdx.view(-1), matches0.view(-1), matches1.view(-1)
        # validmatch = (matches0 >= 0) | (matches1 >= 0)
        # batchIdx, matches0, matches1 = batchIdx[validmatch], matches0[validmatch], matches1[validmatch]
        # matches0[matches0 == -1] = n
        # matches1[matches1 == -1] = m
        # loss_mean = -scores[(batchIdx, matches0, matches1)].mean()
        # loss_mean = nn.l1_loss(loss_mean, jt.float(0.0))

        if not data['return_match']:
            return {'loss': loss}

        with jt.no_grad():
            b, n, m = scores.shape
            # Get the matches with score above "match_threshold".
            indices0, max0 = scores[:, :-1, :-1].argmax(2)
            indices1, max1 = scores[:, :-1, :-1].argmax(1)
            mutual0 = jt.arange(0, n)[None] == indices1.gather(1, indices0)
            mutual1 = jt.arange(0, m)[None] == indices0.gather(1, indices1)
            # zero = scores.new_tensor(0)
            # mscores0 = torch.where(mutual0, max0.values.exp(), zero)
            mscores0 = max0.exp()
            mscores0[mutual0.logical_not()] = 0
            # mscores1 = torch.where(mutual1, mscores0.gather(1, indices1), zero)
            mscores1 = mscores0.gather(1, indices1)
            mscores1[mutual1.logical_not()] = 0
            valid0 = mutual0 & (mscores0 > self.match_threshold)
            valid1 = mutual1 & valid0.gather(1, indices1)
            # indices0 = torch.where(valid0, indices0, indices0.new_tensor(-1))
            # indices1 = torch.where(valid1, indices1, indices1.new_tensor(-1))
            indices0[valid0.logical_not()] = -1
            indices1[valid1.logical_not()] = -1

        return {
            'matches0': indices0,  # use -1 for invalid match
            'matches1': indices1,  # use -1 for invalid match
            'matching_scores0': mscores0,
            'matching_scores1': mscores1,
            'loss': loss,
        }

        # scores big value or small value means confidence? log can't take neg value


# ***************************************************************
# Copyright (c) 2023 Jittor. All Rights Reserved. 
# Maintainers:
#     Guowei Yang <471184555@qq.com>
#     Guoye Yang <498731903@qq.com>
#     Wenyang Zhou <576825820@qq.com>
#     Meng-Hao Guo <guomenghao1997@gmail.com>
#     Dun Liang <randonlang@gmail.com>.
#
# 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import jittor as jt
import numpy as np

class Optimizer(object):
    """ Basic class of Optimizer.

    Example::

        optimizer = nn.SGD(model.parameters(), lr)
        optimizer.step(loss)
    """
    def __init__(self, params, lr, param_sync_iter=10000):
        self.param_groups = []
        self.lr = lr
        self.param_sync_iter = param_sync_iter

        assert len(params) > 0, "Length of parameters should not be zero"
        if not isinstance(params[0], dict):
            params = [{'params': params}]
        for pg in params:
            assert isinstance(pg, dict)
            self.param_groups.append(pg)
        self.n_step = 0
        # __zero_grad is a value for fast determ the grad is zero or not
        # so we can omit 0+x
        self.__zero_grad = True
        self._grad_map = {}

    def add_param_group(self, group):
        self.param_groups.append(group)

    def clip_grad_norm(self, max_norm:float, norm_type:int=2):
        r"""Clips gradient norm of this optimizer.
        The norm is computed over all gradients together.

        Args:
            max_norm (float or int): max norm of the gradients
            norm_type (int): 1-norm or 2-norm

        Example::

            a = jt.ones(2)
            opt = jt.optim.SGD([a], 0.1)

            loss = a*a
            opt.zero_grad()
            opt.backward(loss)

            print(opt.param_groups[0]['grads'][0].norm()) # output: 2.83
            opt.clip_grad_norm(0.01, 2)
            print(opt.param_groups[0]['grads'][0].norm()) # output: 0.01
            
            opt.step()

        """
        if self.__zero_grad: return
        grads = []
        for pg in self.param_groups:
            for p, g in zip(pg["params"], pg["grads"]):
                if p.is_stop_grad(): continue
                grads.append(g.flatten())
        if len(grads) == 0: return
        total_norm = jt.norm(jt.concat(grads), norm_type)
        clip_coef = jt.minimum(max_norm / (total_norm + 1e-6), 1.0)
        for pg in self.param_groups:
            for p, g in zip(pg["params"], pg["grads"]):
                if p.is_stop_grad(): continue
                g.update(g*clip_coef)

    
    @property
    def defaults(self):
        import copy
        exclude = set(("defaults", "pre_step", "step"))
        return copy.deepcopy({ k:v for k, v in self.__dict__.items()
            if k[0] != '_' and k not in exclude and not callable(v) })

    def state_dict(self):
        state = {"defaults": self.defaults}
        return state

    def load_state_dict(self, state):

        def dfs(x):
            if isinstance(x, list):
                for i in range(len(x)):
                    x[i] = dfs(x[i])
            elif isinstance(x, dict):
                for k in x:
                    x[k] = dfs(x[k])
            elif isinstance(x, np.ndarray):
                return jt.array(x).stop_grad()
            elif isinstance(x, jt.Var):
                return x.stop_grad()
            return x
            
        exclude = set(("param_groups",))
        for k, v in state["defaults"].items():
            if k not in exclude:
                setattr(self, k, dfs(v))
        param_groups = dfs(state["defaults"].get('param_groups', None))
        if param_groups is not None:
            exclude = set(("params",))
            for i in range(len(param_groups)):
                for k, v in param_groups[i].items():
                    if k not in exclude:
                        self.param_groups[i][k] = v

        

    def zero_grad(self):
        self.__zero_grad = True

    def backward(self, loss, retain_graph=False):
        '''
        optimize.backward(loss) is used for accumulate multiple step,
        it can be used as following:

        Origin source code ::

        n_iter = 10000
        batch_size = 100
        ...
        for i in range(n_iter):
            ...
            loss = calc_loss()
            optimizer.step(loss)

        Accumulation version ::

        n_iter = 10000
        batch_size = 100
        accumulation_steps = 10
        n_iter *= accumulation_steps
        batch_size //= accumulation_steps
        ...
        for i in range(n_iter):
            ...
            loss = calc_loss()
            # if loss is a mean across batch, we need to divide accumulation_steps
            optimizer.backward(loss / accumulation_steps)
            if (i+1) % accumulation_steps == 0:
                optimizer.step()


        '''
        # clean prev grads
        params = []
        params_has_grad = []
        for pg in self.param_groups:
            for p in pg['params']:
                params.append(p)
                if not p.is_stop_grad():
                    params_has_grad.append(p)

        # sync prev params
        jt.sync(params_has_grad)

        # get gradient
        grads = jt.grad(loss, params_has_grad, retain_graph)

        # sync grads and model if in mpi
        if jt.in_mpi:
            dep = []
            def add_dep(v):
                nonlocal dep
                v._add_dependency(dep)
                dep = [v]

            for g in grads:
                g.assign(g.mpi_all_reduce("mean"))
                add_dep(g._input(0))
            if self.n_step % self.param_sync_iter == 0:
                for p in params:
                    p.assign(p.mpi_broadcast())
                    add_dep(p)
        self.n_step += 1

        # set up grads in param_groups
        pid = 0
        for pg in self.param_groups:
            if "grads" not in pg:
                pg["grads"] = [ jt.zeros_like(p).stop_grad().stop_fuse() for p in pg['params'] ]
            pg_grads = pg["grads"]
            for i, p in enumerate(pg['params']):
                if not p.is_stop_grad():
                    # accumulate grad and stop grad of grad
                    g = grads[pid].stop_grad()
                    if not self.__zero_grad:
                        g = g + pg_grads[i]
                    pg_grads[i].update(g)
                    pid += 1
        self.__zero_grad = False
        
    def pre_step(self, loss, retain_graph=False):
        """ something should be done before step, such as calc gradients, mpi sync, and so on.

        Example::

            class MyOptimizer(Optimizer):
                def step(self, loss):
                    self.pre_step(loss)
                    ...
                    self.post_step()
        """
        if loss is not None:
            self.backward(loss, retain_graph)
        jt.flags.node_order = 1

    def post_step(self):
        """ something should be done before step, such as zero grad, and so on.

        Example::

            class MyOptimizer(Optimizer):
                def step(self, loss):
                    self.pre_step(loss)
                    ...
                    self.post_step()
        """
        jt.flags.node_order = 0
        self.zero_grad()


    def step(self, loss=None, retain_graph=False):
        self.pre_step(loss, retain_graph)
        for pg in self.param_groups:
            lr = pg.get("lr", self.lr)
            for p, g in zip(pg["params"], pg["grads"]):
                if p.is_stop_grad(): continue
                p.update(p - g * lr)
        self.post_step()

    def _build_grad_map(self):
        _grad_map = {}
        for pg in self.param_groups:
            for p, g in zip(pg["params"], pg["grads"]):
                _grad_map[id(p)] = g
        self._grad_map = _grad_map

    def find_grad(self, v:jt.Var) -> jt.Var:
        if id(v) not in self._grad_map:
            self._build_grad_map()
            if id(v) not in self._grad_map:
                raise RuntimeError("This variable is not managed by this optimizer")
        return self._grad_map[id(v)]

def opt_grad(v:jt.Var, opt:Optimizer):
    ''' Get grad of certain variable in optimizer, Example::


    model = Model()
    optimizer = SGD(model.parameters())
    ...
    optimizer.backward(loss)
    
    for p in model.parameters():
        grad = p.opt_grad(optimizer)
    '''
    return opt.find_grad(v)

jt.Var.opt_grad = opt_grad

class SGD(Optimizer):
    """ SGD Optimizer.

    Example::

        optimizer = nn.SGD(model.parameters(), lr, momentum=0.9)
        optimizer.step(loss)
    """
    def __init__(self, params, lr, momentum=0, weight_decay=0, dampening=0, nesterov=False):
        super().__init__(params, lr)
        self.momentum = momentum
        self.weight_decay = weight_decay
        self.dampening = dampening
        self.nesterov = nesterov

        # initialize required arguments
        for pg in self.param_groups:
            values = pg["values"] = []
            for p in pg["params"]:
                values.append(jt.zeros(p.shape, p.dtype).stop_grad())

    def add_param_group(self, group):
        values = group["values"] = []
        for p in group["params"]:
            values.append(jt.zeros(p.shape, p.dtype).stop_grad())
        self.param_groups.append(group)

    def step(self, loss=None, retain_graph=False):
        self.pre_step(loss, retain_graph=False)
        jt.flags.node_order = 1
        for pg in self.param_groups:
            # get arguments from each param_groups
            lr = pg.get("lr", self.lr)
            momentum = pg.get("momentum", self.momentum)
            weight_decay = pg.get("weight_decay", self.weight_decay)
            dampening = pg.get("dampening", self.dampening)
            nesterov = pg.get("nesterov", self.nesterov)

            # optimize main body
            for p, g, v in zip(pg["params"], pg["grads"], pg["values"]):
                if p.is_stop_grad(): continue
                dp = p * weight_decay + g
                v.update(momentum * v + dp * (1 - dampening))
                if nesterov:
                    p.update(p - (dp + momentum * v) * lr)
                else:
                    p.update(p - v * lr)
        self.post_step()

class RMSprop(Optimizer):
    """ RMSprop Optimizer.
    Args:
        params(list): parameters of model.
        lr(float): learning rate.
        eps(float): term added to the denominator to avoid division by zero, default 1e-8.
        alpha(float): smoothing constant, default 0.99.

    Example:
        optimizer = nn.RMSprop(model.parameters(), lr)
        optimizer.step(loss)
    """
    def __init__(self, params, lr=1e-2, eps=1e-8, alpha=0.99):
        super().__init__(params, lr)
        self.eps = eps
        self.alpha = alpha
        
        # initialize required arguments for each param_groups
        for pg in self.param_groups:
            values = pg["values"] = []
            for p in pg["params"]:
                values.append(jt.zeros(p.shape, p.dtype).stop_grad())

    def add_param_group(self, group):
        values = group["values"] = []
        for p in group["params"]:
            values.append(jt.zeros(p.shape, p.dtype).stop_grad())
        self.param_groups.append(group)

    def step(self, loss=None, retain_graph=False):
        self.pre_step(loss, retain_graph)
        for pg in self.param_groups:
            # get arguments from each param_groups
            lr = pg.get("lr", self.lr)
            eps = pg.get("eps", self.eps)
            alpha = pg.get("alpha", self.alpha)
            for p, g, v in zip(pg["params"], pg["grads"], pg["values"]):
                if p.is_stop_grad(): continue
                v.update(alpha * v + (1-alpha) * g * g)
                p.update(p - lr * g / (jt.sqrt(v) + eps))
        self.post_step()

class Adam(Optimizer):
    """ Adam Optimizer.
    
    Example::

        optimizer = nn.Adam(model.parameters(), lr, eps=1e-8, betas=(0.9, 0.999))
        optimizer.step(loss)
    """
    def __init__(self, params, lr, eps=1e-8, betas=(0.9, 0.999), weight_decay=0):
        super().__init__(params, lr)
        self.eps = eps
        self.betas = betas
        self.weight_decay = weight_decay
        # assert weight_decay==0, "weight_decay is not supported yet"
        
        # initialize required arguments for each param_groups
        for pg in self.param_groups:
            values = pg["values"] = []
            m = pg["m"] = []
            for p in pg["params"]:
                values.append(jt.zeros(p.shape, p.dtype).stop_grad())
                m.append(jt.zeros(p.shape, p.dtype).stop_grad())

    def add_param_group(self, group):
        values = group["values"] = []
        m = group["m"] = []
        for p in group["params"]:
            values.append(jt.zeros(p.shape, p.dtype).stop_grad())
            m.append(jt.zeros(p.shape, p.dtype).stop_grad())
        self.param_groups.append(group)

    def step(self, loss=None, retain_graph=False):
        self.pre_step(loss, retain_graph)
        n = float(self.n_step)
        jt.flags.node_order = 1
        for pg in self.param_groups:
            # get arguments from each param_groups
            lr = pg.get("lr", self.lr)
            eps = pg.get("eps", self.eps)
            weight_decay = pg.get("weight_decay", self.weight_decay)
            b0, b1 = pg.get("betas", self.betas)
            for p, g, v, m in zip(pg["params"], pg["grads"], pg["values"], pg["m"]):
                if p.is_stop_grad(): continue
                g = p * weight_decay + g
                m.update(b0 * m + (1-b0) * g)
                v.update(b1 * v + (1-b1) * g * g)
                step_size = lr * jt.sqrt(1-b1**n) / (1-b0 ** n)
                p.update(p - m * step_size / (jt.sqrt(v) + eps))
        self.post_step()


class AdamW(Optimizer):
    """ AdamW Optimizer.
    
    Example::

        optimizer = nn.AdamW(model.parameters(), lr, eps=1e-8, betas=(0.9, 0.999))
        optimizer.step(loss)
    """
    def __init__(self, params, lr, eps=1e-8, betas=(0.9, 0.999), weight_decay=0):
        super().__init__(params, lr)
        self.eps = eps
        self.betas = betas
        self.weight_decay = weight_decay
        # assert weight_decay==0, "weight_decay is not supported yet"
        
        # initialize required arguments for each param_groups
        for pg in self.param_groups:
            values = pg["values"] = []
            m = pg["m"] = []
            for p in pg["params"]:
                values.append(jt.zeros(p.shape, p.dtype).stop_grad())
                m.append(jt.zeros(p.shape, p.dtype).stop_grad())

    def add_param_group(self, group):
        values = group["values"] = []
        m = group["m"] = []
        for p in group["params"]:
            values.append(jt.zeros(p.shape, p.dtype).stop_grad())
            m.append(jt.zeros(p.shape, p.dtype).stop_grad())
        self.param_groups.append(group)

    def step(self, loss=None, retain_graph=False):
        self.pre_step(loss, retain_graph)
        n = float(self.n_step)
        for pg in self.param_groups:
            # get arguments from each param_groups
            lr = pg.get("lr", self.lr)
            eps = pg.get("eps", self.eps)
            weight_decay = pg.get("weight_decay", self.weight_decay)
            b0, b1 = pg.get("betas", self.betas)
            for p, g, v, m in zip(pg["params"], pg["grads"], pg["values"], pg["m"]):
                if p.is_stop_grad(): continue
                p.update(p * (1 - lr * weight_decay))
                bias_correction1 = 1 - b0 ** n
                bias_correction2 = 1 - b1 ** n
                m.update(b0 * m + (1-b0) * g) #exp_avg
                v.update(b1 * v + (1-b1) * g * g) #exp_avg_sq
                denom = jt.sqrt(v) / jt.sqrt(bias_correction2) + eps
                step_size = lr / bias_correction1
                p.update(p - step_size * m / denom)
        self.post_step()
          

class Adan(Optimizer):
    """ Adan Optimizer.
    Adan was proposed in
    Adan: Adaptive Nesterov Momentum Algorithm for
        Faster Optimizing Deep Models[J].arXiv preprint arXiv:2208.06677, 2022.
    https://arxiv.org/abs/2208.06677
    Adan is an efficient optimizer for most DNN frameworks:
    - About 2x fewer computational load than SOTAs
    - Robust to training setting and batch size
    - Easy to Plug-and-play

    Arguments:
        params (iterable): iterable of parameters to optimize or
            dicts defining parameter groups.
        lr (float, optional): learning rate. (default: 1e-3)
        betas (Tuple[float, float, flot], optional): coefficients used for
            first- and second-order moments. (default: (0.98, 0.92, 0.99))
        eps (float, optional): term added to the denominator to improve
            numerical stability. (default: 1e-8)
        weight_decay (float, optional): decoupled weight decay
            (L2 penalty) (default: 0)
        max_grad_norm (float, optional): value used to clip
            global grad norm (default: 0.0 no clip)
    """
    def __init__(self, params, lr=1e-3, betas=(0.98, 0.92, 0.99),
                eps=1e-8, weight_decay=0.0, max_grad_norm=0.0):
        super().__init__(params, lr)
        self.betas = betas
        self.eps = eps
        self.weight_decay = weight_decay
        self.max_grad_norm = max_grad_norm
        
        for pg in self.param_groups:
            pg["m"] = []
            pg["v"] = []
            pg["d"] = []
            pg["pre_grad"] = []
            for p in pg["params"]:
                pg["m"].append(jt.zeros(p.shape, p.dtype).stop_grad())
                pg["v"].append(jt.zeros(p.shape, p.dtype).stop_grad())
                pg["d"].append(jt.zeros(p.shape, p.dtype).stop_grad())
                pg["pre_grad"].append(jt.zeros(p.shape, p.dtype).stop_grad())
                
                
    def add_param_group(self, group):
        group["m"] = []
        group["v"] = []
        group["d"] = []
        group["pre_grad"] = []
        for p in group["params"]:
            group["m"].append(jt.zeros(p.shape, p.dtype).stop_grad())
            group["v"].append(jt.zeros(p.shape, p.dtype).stop_grad())
            group["d"].append(jt.zeros(p.shape, p.dtype).stop_grad())
            group["pre_grad"].append(jt.zeros(p.shape, p.dtype).stop_grad())
        self.param_groups.append(group)

    def step(self, loss=None, retain_graph=False):
        self.pre_step(loss, retain_graph)
        n = float(self.n_step)
        for pg in self.param_groups:
            lr = pg.get("lr", self.lr)
            betas = pg.get("betas", self.betas)
            eps = pg.get("eps", self.eps)
            weight_decay = pg.get("weight_decay", self.weight_decay)
            max_grad_norm = pg.get("max_grad_norm", self.max_grad_norm)
            if max_grad_norm>0: self.clip_grad_norm(max_grad_norm)
            beta1, beta2, beta3 = betas
            
            bias_correction1 = 1 - beta1 ** n
            bias_correction2 = 1 - beta2 ** n
            bias_correction3_sqrt = jt.sqrt(1 - beta3 ** n)
            
            
            step_size_diff = lr * beta2 * bias_correction3_sqrt / bias_correction2
            step_size = lr * bias_correction3_sqrt / bias_correction1
            eps_bias_sqrt = eps * bias_correction3_sqrt
            
            for p, g, m, v, d, pre_g in zip(pg["params"], 
                                            pg["grads"], 
                                            pg["m"], 
                                            pg["v"], 
                                            pg["d"], 
                                            pg["pre_grad"]):
                if p.is_stop_grad(): continue

                if self.n_step>0:
                    pre_g.update(g - pre_g)  # Update pre_g as grad_diff

                
                m.update(beta1 * m + (1 - beta1) * g)
                d.update(beta2 * d + (1 - beta2) * pre_g)  # Use pre_g as grad_diff
                
                pre_g.update(jt.multiply(pre_g, beta2) + g)  # Update pre_g as update (g + beta2 * grad_diff)
                
                v.update(beta3 * v + (1 - beta3) * pre_g * pre_g)  # Use pre_g as update
                
                p.update(p - (step_size * m + step_size_diff * d) / (jt.sqrt(v) + eps_bias_sqrt))
                p.update(p / (1 + lr * weight_decay))
                
                pre_g.update(g)  # Update pre_g for the next iteration
        self.post_step()
        

class LRScheduler:
    def __init__(self,optimizer, last_epoch=-1):
        assert isinstance(optimizer,Optimizer)
        self.optimizer = optimizer
        
        if last_epoch==-1:
            for gp in optimizer.param_groups:
                gp.setdefault('initial_lr',gp.get('lr',optimizer.lr))
        else:
            for gp in optimizer.param_groups:
                assert 'initial_lr' in gp 
        
        self.base_lrs = list(map(lambda group: group['initial_lr'], optimizer.param_groups))
        self.last_epoch = last_epoch
        self.optimizer._step_count = 0
        self._step_count = 0
        self.step()

    def get_lr(self):
        raise NotImplementedError 
    
    def get_last_lr(self):
        return self._last_lr

    def step(self,epoch=None):
        self._step_count += 1

        if epoch is None:
            self.last_epoch += 1
            values = self.get_lr()
        else:
            self.last_epoch = epoch
            values = self.get_lr()

        for i, data in enumerate(zip(self.optimizer.param_groups, values)):
            param_group, lr = data
            param_group['lr'] = lr

        self._last_lr = [group['lr'] for group in self.optimizer.param_groups]


class LambdaLR(LRScheduler):

    def __init__(self, optimizer, lr_lambda, last_epoch=-1):
        if not isinstance(lr_lambda, list) and not isinstance(lr_lambda, tuple):
            self.lr_lambdas = [lr_lambda] * len(optimizer.param_groups)
        else:
            if len(lr_lambda) != len(optimizer.param_groups):
                raise ValueError("Expected {} lr_lambdas, but got {}".format(len(optimizer.param_groups), len(lr_lambda)))

            self.lr_lambdas = list(lr_lambda)
            
        super(LambdaLR, self).__init__(optimizer, last_epoch)
        
        

    def get_lr(self):
        return [base_lr * lmbda(self.last_epoch)
                for lmbda, base_lr in zip(self.lr_lambdas, self.base_lrs)]




# ***************************************************************
# Copyright (c) 2023 Jittor. All Rights Reserved. 
# Maintainers: 
#     Dun Liang <randonlang@gmail.com>. 
#     Xiangli Li <1905692338@qq.com>
#     Jiapeng Zhang <zhangjp20@mails.tsinghua.edu.cn>
# 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************

from cgi import test
import unittest
import jittor as jt
import numpy as np

skip_this_test = False

try:
    jt.dirty_fix_pytorch_runtime_error()
    import torch
except:
    torch = None
    skip_this_test = True

def test_unique_with_torch(input, dim=None):
    jt0, jt1, jt2 = jt.unique(jt.array(input), True, True, dim)
    torch0, torch1, torch2 = torch.unique(torch.tensor(input), True, True, True, dim)
    assert np.allclose(jt0, torch0) and np.allclose(jt1, torch1) and np.allclose(jt2, torch2)


@unittest.skipIf(skip_this_test, "No Torch found")
class TestSparse(unittest.TestCase):

    def test_unique(self):
        test_unique_with_torch(np.array([1, 3, 2, 3, 3, 3], dtype=np.int32))
        test_unique_with_torch(np.array([[1, 3], [2, 3], [1, 2]], dtype=np.int64))

    def test_unique_dim(self):
        test_unique_with_torch(np.array([[1, 3], [2, 3], [1, 3], [2, 3]]), 0)
        test_unique_with_torch(np.array([[1, 3], [2, 3], [1, 3], [2, 3]]), 1)


    @unittest.skipIf(not jt.compiler.has_cuda, "No CUDA found")
    @jt.flag_scope(use_cuda=1)
    def test_unique_cuda(self):
        self.test_unique()

    @unittest.skipIf(not jt.compiler.has_cuda, "No CUDA found")
    @jt.flag_scope(use_cuda=1)
    def test_unique_dim_cuda(self):
        self.test_unique_dim()
    
        
if __name__ == "__main__":
    unittest.main()


# ***************************************************************
# Copyright (c) 2023 Jittor. All Rights Reserved. 
# Maintainers: 
#     Dun Liang <randonlang@gmail.com>. 
# 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import jittor as jt
import unittest

@unittest.skipIf(jt.compile_extern.nccl_ops is None, "no nccl found")
class TestNccl(unittest.TestCase):
    @jt.flag_scope(use_cuda=1)
    def test_nccl(self):
        assert jt.compile_extern.nccl_ops.nccl_test("").data == 123

if __name__ == "__main__":
    unittest.main()



import jittor as jt


def ccl_2d(data_2d):
    ''' 
    2D connected component labelling, original code from https://github.com/DanielPlayne/playne-equivalence-algorithm
    Args:
        [in]param data_2d: binary two-dimensional vector
            type data_2d: jittor array

    Returns:
        [out]result: labeled two-dimensional vector

    Example:
    >>> import jittor as jt
    >>> jt.flags.use_cuda = 1
    >>> import cv2
    >>> import numpy as np
    >>> img = cv2.imread('testImg.png', 0)
    >>> a = img.mean()
    >>> img[img <= a] = 0
    >>> img[img > a] = 1
    >>> img = jt.Var(img)

    >>> result = ccl_2d(img)
    >>> print(jt.unique(result, return_counts=True, return_inverse=True)[0], jt.unique(result, return_counts=True, return_inverse=True)[2])
    >>> cv2.imwrite('testImg_result.png', result.numpy().astype(np.uint8) * 50)
    '''

    data_2d = data_2d.astype(jt.uint32)
    cY = data_2d.shape[0]
    cX = data_2d.shape[1]
    data_2d_copy = data_2d.clone()
    changed = jt.ones([1], dtype=jt.uint32)
    data_2d = data_2d.reshape(cX * cY)
    result = jt.code(data_2d.shape,
                     data_2d.dtype, [data_2d, changed],
                     cuda_header='''
                    @alias(g_image, in0)
                    @alias(g_labels, out)
                    ''',
                     cuda_src=r'''
                    __global__ void init_labels(@ARGS_DEF, const int cX, const int cY) {
                        @PRECALC
                        // Calculate index
                        const unsigned int ix = (blockIdx.x * blockDim.x) + threadIdx.x;
                        const unsigned int iy = (blockIdx.y * blockDim.y) + threadIdx.y;
                        @g_labels(iy*cX + ix) = iy*cX + ix;
                    }

                    __device__ __inline__ unsigned int find_root(@ARGS_DEF, unsigned int label) {
                        // Resolve Label
                        unsigned int next = @g_labels(label);

                        // Follow chain
                        while(label != next) {
                            // Move to next
                            label = next;
                            next = @g_labels(label);
                        }

                        // Return label
                        return label;
                    }

                    __global__ void resolve_labels(@ARGS_DEF, const int cX, const int cY) {
                        @PRECALC
                        // Calculate index
                        const unsigned int id = ((blockIdx.y * blockDim.y) + threadIdx.y) * cX +
                                                ((blockIdx.x * blockDim.x) + threadIdx.x);
                        
                        // Check Thread Range
                        if(id < cX*cY) {
                            // Resolve Label
                            @g_labels(id) = find_root(@ARGS, @g_labels(id));
                        }
                    }

                    __global__ void label_equivalence(@ARGS_DEF, const int cX, const int cY) {
                        @PRECALC
                        // Calculate index
                        const unsigned int ix = (blockIdx.x * blockDim.x) + threadIdx.x;
                        const unsigned int iy = (blockIdx.y * blockDim.y) + threadIdx.y;

                        // Check Thread Range
                        if((ix < cX) && (iy < cY)) {
                            // Get image and label values
                            const unsigned char cyx   = @g_image( iy*cX + ix);
                            
                            // Get neighbour labels
                            const unsigned int lym1x = (iy > 0)    ? @g_labels((iy-1)*cX +   ix) : 0;
                            const unsigned int lyxm1 = (ix > 0)    ? @g_labels(iy    *cX + ix-1) : 0;
                            const unsigned int lyx   =               @g_labels(iy    *cX +   ix);
                            const unsigned int lyxp1 = (ix < cX-1) ? @g_labels(iy    *cX + ix+1) : 0;
                            const unsigned int lyp1x = (iy < cY-1) ? @g_labels((iy+1)*cX +   ix) : 0;

                            const unsigned int lym1xm1 = (iy > 0    && ix > 0   )    ? @g_labels((iy-1)*cX +   ix-1) : 0;
                            const unsigned int lym1xp1 = (iy > 0    && ix < cX-1)    ? @g_labels((iy-1)*cX +   ix+1) : 0;
                            const unsigned int lyp1xm1 = (iy < cY-1 && ix > 0   )    ? @g_labels((iy+1)*cX +   ix-1) : 0;
                            const unsigned int lyp1xp1 = (iy < cY-1 && ix < cX-1)    ? @g_labels((iy+1)*cX +   ix+1) : 0;

                            const bool nym1x = (iy > 0)    ? (cyx == (@g_image((iy-1)*cX +   ix))) : false;
                            const bool nyxm1 = (ix > 0)    ? (cyx == (@g_image(iy    *cX + ix-1))) : false;
                            const bool nyxp1 = (ix < cX-1) ? (cyx == (@g_image(iy    *cX + ix+1))) : false;
                            const bool nyp1x = (iy > cY-1) ? (cyx == (@g_image((iy+1)*cX +   ix))) : false;

                            const bool nym1xm1 = (iy > 0    && ix > 0   )    ? (cyx == (@g_image((iy-1)*cX +   ix-1))) : false;
                            const bool nym1xp1 = (iy > 0    && ix < cX-1)    ? (cyx == (@g_image((iy-1)*cX +   ix+1))) : false;
                            const bool nyp1xm1 = (iy < cY-1 && ix > 0   )    ? (cyx == (@g_image((iy+1)*cX +   ix-1))) : false;
                            const bool nyp1xp1 = (iy < cY-1 && ix < cX-1)    ? (cyx == (@g_image((iy+1)*cX +   ix+1))) : false;

                            // Lowest label
                            unsigned int label = lyx;

                            // Find lowest neighbouring label
                            label = ((nym1x) && (lym1x < label)) ? lym1x : label;
                            label = ((nyxm1) && (lyxm1 < label)) ? lyxm1 : label;
                            label = ((nyxp1) && (lyxp1 < label)) ? lyxp1 : label;
                            label = ((nyp1x) && (lyp1x < label)) ? lyp1x : label;

                            label = ((nym1xm1) && (lym1xm1 < label)) ? lym1xm1 : label;
                            label = ((nym1xp1) && (lym1xp1 < label)) ? lym1xp1 : label;
                            label = ((nyp1xm1) && (lyp1xm1 < label)) ? lyp1xm1 : label;
                            label = ((nyp1xp1) && (lyp1xp1 < label)) ? lyp1xp1 : label;

                            // If labels are different, resolve them
                            if(label < lyx) {
                                // Update label
                                // Nonatomic write may overwrite another label but on average seems to give faster results
                                @g_labels(lyx) = label;

                                // Record the change
                                @in1(0) = 1;
                            }
                        }
                    }
                    ''' + f'''
                    dim3 block(32, 32);
                    const int cX= {cX};
                    const int cY= {cY};''' + '''
                    dim3 grid(ceil(cX/(float)block.x), ceil(cY/(float)block.y));
                    dim3 resolve_block(32, 32);
                    dim3 resolve_grid(ceil(cX/(float)resolve_block.x), ceil(cY/(float)resolve_block.y));
                    
                    // Initialise labels
                    init_labels <<< grid, block >>>(@ARGS, cX, cY);
                    
                    // Resolve the labels
                    resolve_labels <<< resolve_grid, resolve_block >>>(@ARGS, cX, cY);

                    // Changed Flag
                    int32 changed = 1;
                    
                    // While labels have changed
                    while(changed) {
                        // Copy changed to device
                        cudaMemsetAsync(in1_p, 0, 4);
                        
                        // Label image
                        label_equivalence <<< grid, block >>>(@ARGS, cX, cY);

                        // Copy changed flag to host
                        cudaMemcpy(&changed, in1_p, sizeof(int32), cudaMemcpyDeviceToHost);
                        
                        // Resolve the labels
                        resolve_labels <<< resolve_grid, resolve_block>>>(@ARGS, cX, cY);
                    }
                    ''')
    result = result.reshape((cY, cX)) * data_2d_copy
    value = jt.unique(result)
    value = value[value != 0]

    map_result = jt.zeros((int(value.max().numpy()[0]) + 1), dtype=jt.uint32)
    map_result[value] = jt.index(value.shape)[0] + 1
    result = map_result[result]

    return result



# ***************************************************************
# Copyright (c) 2023 Jittor. All Rights Reserved. 
# Maintainers:
#     Meng-Hao Guo <guomenghao1997@gmail.com>
#     Dun Liang <randonlang@gmail.com>.
# 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************

import jittor as jt
import numpy as np
from collections.abc import Sequence, Mapping
from PIL import Image
import time

def get_random_list(n):
    return list(np.random.permutation(range(n)))

def get_order_list(n):
    return [i for i in range(n)]


def collate_batch(batch):
    r"""Puts each data field into a tensor with outer dimension batch size"""
    real_size = len(batch)
    elem = batch[0]
    elem_type = type(elem)
    if isinstance(elem, jt.Var):
        temp_data = jt.stack([data for data in batch], 0)
        return temp_data
    if elem_type is np.ndarray:
        temp_data = np.stack([data for data in batch], 0)
        return temp_data
    elif np.issubdtype(elem_type, np.integer):
        return np.int32(batch)
    elif isinstance(elem, int):
        return np.int32(batch)
    elif isinstance(elem, float):
        return np.float32(batch)
    elif isinstance(elem, str):
        return batch
    elif isinstance(elem, Mapping):
        return {key: collate_batch([d[key] for d in batch]) for key in elem}
    elif isinstance(elem, tuple):
        transposed = zip(*batch)
        return tuple(collate_batch(samples) for samples in transposed)
    elif isinstance(elem, Sequence):
        transposed = zip(*batch)
        return [collate_batch(samples) for samples in transposed]
    elif isinstance(elem, Image.Image):
        temp_data = np.stack([np.array(data) for data in batch], 0)
        return temp_data
    else:
        raise TypeError(f"Not support type <{elem_type.__name__}>")

class HookTimer:
    def __init__(self, obj, attr):
        self.origin = getattr(obj, attr)
        self.duration = 0.0
        setattr(obj, attr, self)

    def __call__(self, *args, **kw):
        start = time.time()
        rt = self.origin(*args, **kw)
        self.duration += time.time() - start
        return rt




# ***************************************************************
# Copyright (c) 2023 Jittor. All Rights Reserved. 
# Maintainers: 
#     Wenyang Zhou <576825820@qq.com>
#     Dun Liang <randonlang@gmail.com>. 
# 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import jittor as jt
import numpy as np
import jittor.distributions as jd

skip_this_test = False
try:
    jt.dirty_fix_pytorch_runtime_error()
    import torch
except:
    torch = None
    skip_this_test = True


class TestOneHot(unittest.TestCase):
    def test_presum(self):
        a = jt.array([[1,2,3,4]])
        b = jd.simple_presum(a)
        assert (b.data == [[0,1,3,6,10]]).all()

    @unittest.skipIf(skip_this_test, "No Torch Found")
    def test_one_hot(self):
        a = jd.OneHotCategorical(jt.array([0.25, 0.25, 0.25, 0.25]))
        x = a.sample().numpy()
        for i in range(1000):
            x += a.sample().numpy()
        assert (x > 200).all()
        y = a.sample([2,3])
        y.sync()
        assert y.shape == [2,3,4]
        probs,probs2 = np.random.uniform(0,1,(10)), np.random.uniform(0,1,(10))
        probs,probs2 = probs / probs.sum(),probs2 / probs2.sum()

        jc, jc2 = jd.OneHotCategorical(jt.array(probs)),jd.OneHotCategorical(jt.array(probs2))
        tc, tc2 = torch.distributions.OneHotCategorical(torch.tensor(probs)),torch.distributions.OneHotCategorical(torch.tensor(probs2))
        assert np.allclose(jc.entropy().data,tc.entropy().numpy())
        x = np.zeros((4,10))
        for _ in range(4):
            nx = np.random.randint(0,9)    
            x[_,nx] = 1
        np.testing.assert_allclose(jc.log_prob(jt.array(x)),tc.log_prob(torch.tensor(x)), atol=1e-5)
        assert np.allclose(jd.kl_divergence(jc,jc2),torch.distributions.kl_divergence(tc,tc2))

    def test_cate(self):
        a = jd.Categorical(jt.array([0.25, 0.25, 0.25, 0.25]))
        x =np.array([0,0,0,0])
        for i in range(1000):
            x[a.sample().item()]+=1
        assert (x > 200).all()
        y = a.sample([2,3])
        y.sync()
        assert y.shape == [2,3]
        
    @unittest.skipIf(skip_this_test, "No Torch Found")
    def test_normal(self):
        for _ in range(4):
            mu = np.random.uniform(-1,1)
            sigma = np.random.uniform(0,2)
            jn = jd.Normal(mu,sigma)
            tn = torch.distributions.Normal(mu,sigma)
            assert np.allclose(jn.entropy().data,tn.entropy().numpy())
            x = np.random.uniform(-1,1)
            np.testing.assert_allclose(jn.log_prob(x),tn.log_prob(torch.tensor(x)))
            mu2 = np.random.uniform(-1,1)
            sigma2 = np.random.uniform(0,2)
            jn2 = jd.Normal(mu2,sigma2)
            tn2 = torch.distributions.Normal(mu2,sigma2)
            assert np.allclose(jd.kl_divergence(jn,jn2).data,torch.distributions.kl_divergence(tn,tn2).numpy())

    @unittest.skipIf(skip_this_test, "No Torch Found")
    def test_categorical1(self):
        for _ in range(4):
            probs,probs2 = np.random.uniform(0,1,(10)), np.random.uniform(0,1,(10))
            probs,probs2 = probs / probs.sum(),probs2 / probs2.sum()
            jc, jc2 = jd.Categorical(jt.array(probs)),jd.Categorical(jt.array(probs2))
            tc, tc2 = torch.distributions.Categorical(torch.tensor(probs)),torch.distributions.Categorical(torch.tensor(probs2))
            assert np.allclose(jc.entropy().data, tc.entropy().numpy()), (jc.entropy().data, tc.entropy().numpy())
            x = np.random.randint(0,10,(4))
            np.testing.assert_allclose(jc.log_prob(x), tc.log_prob(torch.tensor(x)), atol=1e-5)
            assert np.allclose(jd.kl_divergence(jc,jc2),torch.distributions.kl_divergence(tc,tc2))

    @unittest.skipIf(skip_this_test, "No Torch Found")
    def test_categorical2(self):
        def check(prob_shape, sample_shape):
            for _ in range(4):
                probs,probs2 = np.random.uniform(0,1,prob_shape), np.random.uniform(0,1, prob_shape)

                jc, jc2 = jd.Categorical(jt.array(probs)),jd.Categorical(jt.array(probs2))
                tc, tc2 = torch.distributions.Categorical(torch.tensor(probs)),torch.distributions.Categorical(torch.tensor(probs2))
                assert np.allclose(jc.entropy().data, tc.entropy().numpy()), (jc.entropy().data, tc.entropy().numpy())
                x1 = jc.sample(sample_shape)
                x2 = tc.sample(sample_shape)
                assert tuple(x1.shape) == tuple(x2.shape)
                x = np.random.randint(0,prob_shape[-1], tuple(x1.shape))
                np.testing.assert_allclose(jc.log_prob(x), tc.log_prob(torch.tensor(x)), atol=1e-5)
                np.testing.assert_allclose(jd.kl_divergence(jc,jc2), torch.distributions.kl_divergence(tc,tc2), atol=1e-5)
        check((10,), (4,))
        check((2,3), (4,))
        check((3,4,5,6), (2,))

    @unittest.skipIf(skip_this_test, "No Torch Found")
    def test_one_hot_categorical2(self):
        def check(prob_shape, sample_shape):
            for _ in range(4):
                probs,probs2 = np.random.uniform(0,1,prob_shape), np.random.uniform(0,1, prob_shape)

                jc, jc2 = jd.OneHotCategorical(jt.array(probs)),jd.OneHotCategorical(jt.array(probs2))
                tc, tc2 = torch.distributions.OneHotCategorical(torch.tensor(probs)),torch.distributions.OneHotCategorical(torch.tensor(probs2))
                assert np.allclose(jc.entropy().data, tc.entropy().numpy()), (jc.entropy().data, tc.entropy().numpy())
                x1 = jc.sample(sample_shape)
                x2 = tc.sample(sample_shape)
                assert tuple(x1.shape) == tuple(x2.shape)
                x = np.random.randint(0,prob_shape[-1], tuple(x1.shape))
                np.testing.assert_allclose(jc.log_prob(x), tc.log_prob(torch.tensor(x)), atol=1e-5)
                np.testing.assert_allclose(jd.kl_divergence(jc,jc2), torch.distributions.kl_divergence(tc,tc2), atol=1e-5)
        check((10,), (4,))
        check((2,3), (4,))
        check((3,4,5,6), (2,))
            
    @unittest.skipIf(skip_this_test, "No Torch Found")
    def test_uniform(self):
        for _ in range(4):
            low, low2 = np.random.randint(-1,2), np.random.randint(-1,2)
            leng, leng2 = np.random.uniform(0,2), np.random.uniform(0,2)
            high, high2 = low + leng, low2 + leng2
            ju, ju2 = jd.Uniform(low,high),jd.Uniform(low2,high2)
            tu, tu2 = torch.distributions.Uniform(low,high),torch.distributions.Uniform(low2,high2)
            assert np.allclose(ju.entropy().data,tu.entropy().numpy())
            x = np.random.uniform(low,high)
            assert np.allclose(ju.log_prob(x),tu.log_prob(torch.tensor(x)))
            assert np.allclose(jd.kl_divergence(ju,ju2),torch.distributions.kl_divergence(tu,tu2))
    
    @unittest.skipIf(skip_this_test, "No Torch Found")
    def test_geometric(self):
        for _ in range(4):
            prob, prob2 = np.random.uniform(0,1), np.random.uniform(0,1)
            jg, jg2 = jd.Geometric(prob),jd.Geometric(prob2)
            tg, tg2 = torch.distributions.Geometric(prob),torch.distributions.Geometric(prob2)
            np.testing.assert_allclose(jg.entropy().data,tg.entropy().numpy(), atol=1e-4)
            x = np.random.randint(1,10)
            np.testing.assert_allclose(jg.log_prob(x),tg.log_prob(torch.tensor(x)), atol=1e-4)
            # print(jd.kl_divergence(jg,jg2),torch.distributions.kl_divergence(tg,tg2))
            np.testing.assert_allclose(jd.kl_divergence(jg,jg2),torch.distributions.kl_divergence(tg,tg2), atol=1e-4)

if __name__ == "__main__":
    unittest.main()